# Copyright (C) 2022 The Electrum developers
# Distributed under the MIT software license, see the accompanying
# file LICENCE or http://www.opensource.org/licenses/mit-license.php

from typing import Optional, TYPE_CHECKING

from PyQt6.QtGui import QFont, QCursor, QMouseEvent
from PyQt6.QtCore import Qt, QSize
from PyQt6.QtWidgets import (QLabel, QVBoxLayout, QGridLayout, QTextEdit,
                             QHBoxLayout, QPushButton, QWidget, QSizePolicy, QFrame)

from electrum.i18n import _
from electrum.util import InvoiceError, ChoiceItem
from electrum.invoices import pr_expiration_values
from electrum.logging import Logger

from .amountedit import AmountEdit, BTCAmountEdit, SizedFreezableLineEdit
from .qrcodewidget import QRCodeWidget
from .util import read_QIcon, WWLabel, MessageBoxMixin, MONOSPACE_FONT, get_icon_qrcode

if TYPE_CHECKING:
    from .main_window import ElectrumWindow
    from electrum.wallet import Request


class ReceiveTab(QWidget, MessageBoxMixin, Logger):

    # strings updated by update_current_request
    addr = ''
    lnaddr = ''
    URI = ''
    address_help = ''
    URI_help = ''
    ln_help = ''

    def __init__(self, window: 'ElectrumWindow'):
        QWidget.__init__(self, window)
        Logger.__init__(self)

        self.window = window
        self.wallet = window.wallet
        self.fx = window.fx
        self.config = window.config

        # A 4-column grid layout.  All the stretch is in the last column.
        # The exchange rate plugin adds a fiat widget in column 2
        self.receive_grid = grid = QGridLayout()
        grid.setSpacing(8)
        grid.setColumnStretch(3, 1)

        self.receive_message_e = SizedFreezableLineEdit(width=400)
        grid.addWidget(QLabel(_('Description')), 0, 0)
        grid.addWidget(self.receive_message_e, 0, 1, 1, 4)

        self.receive_amount_e = BTCAmountEdit(self.window.get_decimal_point)
        grid.addWidget(QLabel(_('Requested amount')), 1, 0)
        grid.addWidget(self.receive_amount_e, 1, 1)

        self.fiat_receive_e = AmountEdit(self.fx.get_currency if self.fx else '')
        if not self.fx or not self.fx.is_enabled():
            self.fiat_receive_e.setVisible(False)
        grid.addWidget(self.fiat_receive_e, 1, 2, Qt.AlignmentFlag.AlignLeft)

        self.window.connect_fields(self.receive_amount_e, self.fiat_receive_e)

        self.expiry_button = QPushButton('')
        self.expiry_button.clicked.connect(self.expiry_dialog)
        grid.addWidget(QLabel(_('Expiry')), 2, 0)
        grid.addWidget(self.expiry_button, 2, 1)

        self.clear_invoice_button = QPushButton(_('Clear'))
        self.clear_invoice_button.clicked.connect(self.do_clear)
        text = _('Onchain') if self.wallet.has_lightning() else _('Request')
        self.create_onchain_invoice_button = QPushButton(text)
        self.create_onchain_invoice_button.setIcon(read_QIcon("bitcoin.png"))
        self.create_onchain_invoice_button.clicked.connect(lambda: self.create_invoice(False))
        self.create_lightning_invoice_button = QPushButton(_('Lightning'))
        self.create_lightning_invoice_button.setIcon(read_QIcon("lightning.png"))
        self.create_lightning_invoice_button.clicked.connect(lambda: self.create_invoice(True))
        self.create_lightning_invoice_button.setVisible(self.wallet.has_lightning())

        self.receive_buttons = buttons = QHBoxLayout()
        buttons.addWidget(self.clear_invoice_button)
        buttons.addStretch(1)
        buttons.addWidget(self.create_onchain_invoice_button)
        buttons.addWidget(self.create_lightning_invoice_button)
        grid.addLayout(buttons, 4, 1, 1, -1)

        self.receive_e = QTextEdit()
        self.receive_e.setFont(QFont(MONOSPACE_FONT))
        self.receive_e.setReadOnly(True)
        self.receive_e.setContextMenuPolicy(Qt.ContextMenuPolicy.NoContextMenu)
        self.receive_e.setTextInteractionFlags(Qt.TextInteractionFlag.NoTextInteraction)
        self.receive_e.textChanged.connect(self.update_receive_widgets)

        self.receive_qr = QRCodeWidget(manual_size=True)

        self.receive_help_text = WWLabel('')
        self.receive_help_text.setLayout(QHBoxLayout())
        self.receive_rebalance_button = QPushButton('Rebalance')
        self.receive_rebalance_button.suggestion = None
        self.receive_zeroconf_button = QPushButton(_('Accept'))
        self.receive_zeroconf_button.clicked.connect(self.on_accept_zeroconf)

        self.previous_request = None  # type: Optional['Request']
        self.confirmed_zeroconf_for_this_request = False  # type: bool

        def on_receive_rebalance():
            if self.receive_rebalance_button.suggestion:
                chan1, chan2, delta = self.receive_rebalance_button.suggestion
                self.window.rebalance_dialog(chan1, chan2, amount_sat=delta)
        self.receive_rebalance_button.clicked.connect(on_receive_rebalance)
        self.receive_swap_button = QPushButton('Swap')
        self.receive_swap_button.suggestion = None

        def on_receive_swap():
            if self.receive_swap_button.suggestion:
                chan, swap_recv_amount_sat = self.receive_swap_button.suggestion
                self.window.run_swap_dialog(is_reverse=True, recv_amount_sat_or_max=swap_recv_amount_sat, channels=[chan])
        self.receive_swap_button.clicked.connect(on_receive_swap)
        buttons = QHBoxLayout()
        buttons.addWidget(self.receive_rebalance_button)
        buttons.addWidget(self.receive_swap_button)
        buttons.addWidget(self.receive_zeroconf_button)
        vbox = QVBoxLayout()
        vbox.addWidget(self.receive_help_text)
        vbox.addLayout(buttons)
        self.receive_help_widget = FramedWidget()
        self.receive_help_widget.setVisible(False)
        self.receive_help_widget.setLayout(vbox)

        self.receive_widget = ReceiveWidget(
            self, self.receive_e, self.receive_qr, self.receive_help_widget)
        #self.receive_widget.mouseReleaseEvent = lambda x: self.toggle_receive_qr()

        receive_widget_sp = QSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        receive_widget_sp.setRetainSizeWhenHidden(True)
        self.receive_widget.setSizePolicy(receive_widget_sp)
        self.receive_widget.setVisible(False)

        self.receive_requests_label = QLabel(_('Requests'))
        # with QDarkStyle, this label may partially cover the qrcode widget.
        # setMaximumWidth prevents that
        self.receive_requests_label.setMaximumWidth(400)
        from .request_list import RequestList
        self.request_list = RequestList(self)
        # toolbar
        self.toolbar, menu = self.request_list.create_toolbar_with_menu('')

        self.toggle_qr_button = QPushButton('')
        self.toggle_qr_button.setIcon(get_icon_qrcode())
        self.toggle_qr_button.setToolTip(_('Switch between text and QR code view'))
        self.toggle_qr_button.clicked.connect(self.toggle_receive_qr)
        self.toggle_qr_button.setEnabled(False)
        self.toolbar.insertWidget(2, self.toggle_qr_button)

        # menu
        self.qr_menu_action = menu.addToggle(_("Show detached QR code window"), self.window.toggle_qr_window)
        menu.addAction(_("Import requests"), self.window.import_requests)
        menu.addAction(_("Export requests"), self.window.export_requests)
        menu.addAction(_("Delete expired requests"), self.request_list.delete_expired_requests)
        self.toolbar_menu = menu

        # layout
        vbox_g = QVBoxLayout()
        vbox_g.addLayout(grid)
        vbox_g.addStretch()
        hbox = QHBoxLayout()
        hbox.addLayout(vbox_g)
        hbox.addStretch()
        hbox.addWidget(self.receive_widget, 1)

        self.searchable_list = self.request_list
        vbox = QVBoxLayout(self)
        vbox.addLayout(self.toolbar)
        vbox.addLayout(hbox)
        vbox.addStretch()
        vbox.addWidget(self.receive_requests_label)
        vbox.addWidget(self.request_list)
        vbox.setStretchFactor(hbox, 40)
        vbox.setStretchFactor(self.request_list, 60)
        self.request_list.update()  # after parented and put into a layout, can update without flickering
        self.update_expiry_text()

    def update_expiry_text(self):
        expiry = self.config.WALLET_PAYREQ_EXPIRY_SECONDS
        text = pr_expiration_values()[expiry]
        self.expiry_button.setText(text)

    def expiry_dialog(self):
        msg = ''.join([
            _('Expiration period of your request.'), ' ',
            _('This information is seen by the recipient if you send them a signed payment request.'),
            '\n\n',
            _('For on-chain requests, the address gets reserved until expiration. After that, it might get reused.'), ' ',
            _('The bitcoin address never expires and will always be part of this electrum wallet.'), ' ',
            _('You can reuse a bitcoin address any number of times but it is not good for your privacy.'),
            '\n\n',
            _('For Lightning requests, payments will not be accepted after the expiration.'),
        ])
        expiry = self.config.WALLET_PAYREQ_EXPIRY_SECONDS
        choices = [ChoiceItem(key=exptime, label=label)
                   for (exptime, label) in pr_expiration_values().items()]
        v = self.window.query_choice(msg, choices, title=_('Expiry'), default_key=expiry)
        if v is None:
            return
        self.config.WALLET_PAYREQ_EXPIRY_SECONDS = v
        self.update_expiry_text()

    def on_tab_changed(self):
        text, data, help_text, title = self.get_tab_data()
        self.window.do_copy(text, title=title)
        self.update_receive_qr_window()

    def do_copy(self, e: 'QMouseEvent'):
        if e.button() != Qt.MouseButton.LeftButton:
            return
        text, data, help_text, title = self.get_tab_data()
        self.window.do_copy(text, title=title)

    def toggle_receive_qr(self):
        b = not self.config.GUI_QT_RECEIVE_TAB_QR_VISIBLE
        self.config.GUI_QT_RECEIVE_TAB_QR_VISIBLE = b
        self.update_receive_widgets()

    def update_receive_widgets(self):
        b = self.config.GUI_QT_RECEIVE_TAB_QR_VISIBLE
        self.receive_widget.update_visibility(b, bool(self.receive_help_text.text()))

    def update_current_request(self):
        if len(self.request_list.selectionModel().selectedRows(0)) > 1:
            key = None
        else:
            key = self.request_list.get_current_key()
        req = self.wallet.get_request(key) if key else None
        if req != self.previous_request:
            self.previous_request = req
            self.confirmed_zeroconf_for_this_request = False
        if req is None:
            self.receive_e.setText('')
            self.addr = self.URI = self.lnaddr = ''
            self.address_help = self.URI_help = self.ln_help = ''
            return
        help_texts = self.wallet.get_help_texts_for_receive_request(req)
        self.addr = (req.get_address() or '') if not help_texts.address_is_error else ''
        self.URI = (self.wallet.get_request_URI(req) or '') if not help_texts.URI_is_error else ''
        self.lnaddr = self.wallet.get_bolt11_invoice(req) if not help_texts.ln_is_error else ''
        self.address_help = help_texts.address_help
        self.URI_help = help_texts.URI_help
        self.ln_help = help_texts.ln_help
        can_rebalance = help_texts.can_rebalance()
        can_swap = help_texts.can_swap()
        can_zeroconf = help_texts.can_zeroconf() if not self.confirmed_zeroconf_for_this_request else False
        self.receive_rebalance_button.suggestion = help_texts.ln_rebalance_suggestion
        self.receive_swap_button.suggestion = help_texts.ln_swap_suggestion
        self.receive_rebalance_button.setVisible(can_rebalance)
        self.receive_swap_button.setVisible(can_swap)
        self.receive_rebalance_button.setEnabled(can_rebalance and self.window.num_tasks() == 0)
        self.receive_swap_button.setEnabled(can_swap and self.window.num_tasks() == 0)
        self.receive_zeroconf_button.setVisible(can_zeroconf)
        self.receive_zeroconf_button.setEnabled(can_zeroconf)
        text, data, help_text, title = self.get_tab_data()
        if self.confirmed_zeroconf_for_this_request and help_texts.can_zeroconf():
            help_text = ''
        # set help before receive_e so we don't flicker from qr to help
        self.receive_help_text.setText(help_text)
        self.receive_e.setText(text)
        self.receive_qr.setData(data)
        for w in [self.receive_e, self.receive_qr]:
            w.setEnabled(bool(text) and (not help_text or can_zeroconf))
            w.setToolTip(help_text)
        # macOS hack (similar to #4777)
        self.receive_e.repaint()
        # always show
        self.receive_widget.setVisible(True)
        self.toggle_qr_button.setEnabled(True)
        self.update_receive_qr_window()

    def on_accept_zeroconf(self):
        self.receive_zeroconf_button.setVisible(False)
        self.confirmed_zeroconf_for_this_request = True
        self.receive_help_text.setText('')
        self.update_receive_widgets()

    def get_tab_data(self):
        if self.URI:
            out = self.URI, self.URI, self.URI_help, _('Bitcoin URI')
        elif self.addr:
            out = self.addr, self.addr, self.address_help, _('Address')
        else:
            # encode lightning invoices as uppercase so QR encoding can use
            # alphanumeric mode; resulting in smaller QR codes
            out = self.lnaddr, self.lnaddr.upper(), self.ln_help, _('Lightning Request')
        return out

    def update_receive_qr_window(self):
        if self.window.qr_window and self.window.qr_window.isVisible():
            text, data, help_text, title = self.get_tab_data()
            self.window.qr_window.qrw.setData(data)

    def create_invoice(self, is_lightning: bool):
        amount_sat = self.receive_amount_e.get_amount()
        message = self.receive_message_e.text()
        expiry = self.config.WALLET_PAYREQ_EXPIRY_SECONDS
        if is_lightning:
            address = None
        else:
            if amount_sat and amount_sat < self.wallet.dust_threshold():
                self.show_error(_('Amount too small to be received onchain'))
                return
            address = self.get_bitcoin_address_for_request(amount_sat)
            if not address:
                return
            self.window.address_list.update()

        # generate even if we cannot receive
        try:
            key = self.wallet.create_request(amount_sat, message, expiry, address)
        except InvoiceError as e:
            self.show_error(_('Error creating payment request') + ':\n' + str(e))
            return
        except Exception as e:
            self.logger.exception('Error adding payment request')
            self.show_error(_('Error adding payment request') + ':\n' + repr(e))
            return
        assert key is not None
        self.window.address_list.refresh_all()
        self.request_list.update()
        self.request_list.set_current_key(key)
        # clear request fields
        self.receive_amount_e.setText('')
        self.receive_message_e.setText('')
        # copy current tab to clipboard
        self.on_tab_changed()

    def get_bitcoin_address_for_request(self, amount) -> Optional[str]:
        addr = self.wallet.get_unused_address()
        if addr is None:
            if not self.wallet.is_deterministic():  # imported wallet
                msg = [
                    _('No more addresses in your wallet.'), ' ',
                    _('You are using a non-deterministic wallet, which cannot create new addresses.'), ' ',
                    _('If you want to create new addresses, use a deterministic wallet instead.'), '\n\n',
                    _('Creating a new payment request will reuse one of your addresses and overwrite an existing request. Continue anyway?'),
                   ]
                if not self.question(''.join(msg)):
                    return
                addr = self.wallet.get_receiving_address()
            else:  # deterministic wallet
                if not self.question(_("Warning: The next address will not be recovered automatically if you restore your wallet from seed; you may need to add it manually.\n\nThis occurs because you have too many unused addresses in your wallet. To avoid this situation, use the existing addresses first.\n\nCreate anyway?")):
                    return
                addr = self.wallet.create_new_address(False)
        return addr

    def do_clear(self):
        self.receive_e.setText('')
        self.addr = self.URI = self.lnaddr = ''
        self.address_help = self.URI_help = self.ln_help = ''
        self.receive_widget.setVisible(False)
        self.toggle_qr_button.setEnabled(False)
        self.receive_message_e.setText('')
        self.receive_amount_e.setAmount(None)
        self.request_list.clearSelection()


class ReceiveWidget(QWidget):
    min_size = QSize(200, 200)

    def __init__(self, receive_tab: 'ReceiveTab', textedit: QWidget, qr: QWidget, help_widget: QWidget):
        QWidget.__init__(self)
        self.textedit = textedit
        self.qr = qr
        self.help_widget = help_widget
        self.setMinimumSize(self.min_size)

        for w in [textedit, qr]:
            w.mousePressEvent = receive_tab.do_copy
            w.setCursor(QCursor(Qt.CursorShape.PointingHandCursor))

        textedit.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        if isinstance(help_widget, QLabel):
            help_widget.setFrameStyle(QFrame.Shape.StyledPanel)
            help_widget.setStyleSheet("QLabel {border:1px solid gray; border-radius:2px; }")

        hbox = QHBoxLayout()
        hbox.addStretch()
        hbox.addWidget(textedit)
        hbox.addWidget(help_widget)
        hbox.addWidget(qr)

        vbox = QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addStretch()

        self.setLayout(vbox)

    def update_visibility(self, is_qr: bool, show_help: bool):
        if str(self.textedit.toPlainText()) and not show_help:
            self.help_widget.setVisible(False)
            self.textedit.setVisible(not is_qr)
            self.qr.setVisible(is_qr)
        else:
            self.show_help()

    def show_help(self):
        self.help_widget.setVisible(True)
        self.textedit.setVisible(False)
        self.qr.setVisible(False)

    def resizeEvent(self, e):
        # keep square aspect ratio when resized
        size = e.size()
        margin = 10
        x = min(size.height(), size.width()) - margin
        for w in [self.textedit, self.qr, self.help_widget]:
            w.setFixedWidth(x)
            w.setFixedHeight(x)
        return super().resizeEvent(e)


class FramedWidget(QFrame):
    def __init__(self):
        QFrame.__init__(self)
        self.setFrameStyle(QFrame.Shape.StyledPanel)
        self.setStyleSheet("FramedWidget {border:1px solid gray; border-radius:2px; }")



# Electrum - lightweight Bitcoin client
# Copyright (C) 2016  The Electrum developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unicodedata import normalize
import hashlib
import re
import copy
from typing import Tuple, TYPE_CHECKING, Union, Sequence, Optional, Dict, List, NamedTuple, Any, Type
from functools import wraps
from abc import ABC, abstractmethod

import electrum_ecc as ecc
from electrum_ecc import string_to_number

from . import bitcoin, constants, bip32
from .bitcoin import deserialize_privkey, serialize_privkey, BaseDecodeError
from .transaction import Transaction, PartialTransaction, PartialTxInput, PartialTxOutput, TxInput
from .bip32 import (convert_bip32_strpath_to_intpath, BIP32_PRIME,
                    is_xpub, is_xprv, BIP32Node, normalize_bip32_derivation,
                    convert_bip32_intpath_to_strpath, is_xkey_consistent_with_key_origin_info,
                    KeyOriginInfo)
from .descriptor import PubkeyProvider
from . import crypto
from .crypto import (pw_decode, pw_encode, sha256, sha256d, PW_HASH_VERSION_LATEST,
                     SUPPORTED_PW_HASH_VERSIONS, UnsupportedPasswordHashVersion, hash_160,
                     CiphertextFormatError)
from .util import (InvalidPassword, WalletFileException,
                   BitcoinException, bfh, inv_dict, is_hex_str)
from .mnemonic import Mnemonic, Wordlist, calc_seed_type, is_seed
from .plugin import run_hook
from .logging import Logger
from .lrucache import LRUCache

if TYPE_CHECKING:
    from .gui.common_qt.util import TaskThread
    from .hw_wallet import HW_PluginBase, HardwareClientBase, HardwareHandlerBase
    from .wallet_db import WalletDB
    from .plugin import Device


class CannotDerivePubkey(Exception): pass
class ScriptTypeNotSupported(Exception): pass


def also_test_none_password(check_password_fn):
    """Decorator for check_password, simply to give a friendlier exception if
    check_password(x) is called on a keystore that does not have a password set.
    """
    @wraps(check_password_fn)
    def wrapper(self: 'Software_KeyStore', *args):
        password = args[0]
        try:
            return check_password_fn(self, password)
        except (CiphertextFormatError, InvalidPassword) as e:
            if password is not None:
                try:
                    check_password_fn(self, None)
                except Exception:
                    pass
                else:
                    raise InvalidPassword("password given but keystore has no password") from e
            raise
    return wrapper


class KeyStore(Logger, ABC):
    type: str

    def __init__(self):
        Logger.__init__(self)
        self.is_requesting_to_be_rewritten_to_wallet_file = False  # type: bool

    def has_seed(self) -> bool:
        return False

    def is_watching_only(self) -> bool:
        return False

    def can_import(self) -> bool:
        return False

    def get_type_text(self) -> str:
        return f'{self.type}'

    @abstractmethod
    def may_have_password(self) -> bool:
        """Returns whether the keystore can be encrypted with a password."""
        pass

    def _get_tx_derivations(self, tx: 'PartialTransaction') -> Dict[bytes, Union[Sequence[int], str]]:
        keypairs = {}
        for txin in tx.inputs():
            keypairs.update(self._get_txin_derivations(txin))
        return keypairs

    def _get_txin_derivations(self, txin: 'PartialTxInput') -> Dict[bytes, Union[Sequence[int], str]]:
        if txin.is_complete():
            return {}
        keypairs = {}
        for pubkey in txin.pubkeys:
            if pubkey in txin.sigs_ecdsa:
                # this pubkey already signed
                continue
            derivation = self.get_pubkey_derivation(pubkey, txin)
            if not derivation:
                continue
            keypairs[pubkey] = derivation
        return keypairs

    def can_sign(self, tx: 'Transaction', *, ignore_watching_only: bool = False) -> bool:
        """Returns whether this keystore could sign *something* in this tx."""
        if not ignore_watching_only and self.is_watching_only():
            return False
        if not isinstance(tx, PartialTransaction):
            return False
        return bool(self._get_tx_derivations(tx))

    def can_sign_txin(self, txin: 'TxInput', *, ignore_watching_only: bool = False) -> bool:
        """Returns whether this keystore could sign this txin."""
        if not ignore_watching_only and self.is_watching_only():
            return False
        if not isinstance(txin, PartialTxInput):
            return False
        return bool(self._get_txin_derivations(txin))

    def ready_to_sign(self) -> bool:
        return not self.is_watching_only()

    @abstractmethod
    def dump(self) -> dict[str, Any]:
        pass

    @abstractmethod
    def is_deterministic(self) -> bool:
        pass

    @abstractmethod
    def sign_message(
            self,
            sequence: 'AddressIndexGeneric',
            message: str,
            password,
            *,
            script_type: Optional[str] = None,
    ) -> bytes:
        pass

    @abstractmethod
    def decrypt_message(self, sequence: 'AddressIndexGeneric', message, password) -> bytes:
        pass

    @abstractmethod
    def sign_transaction(self, tx: 'PartialTransaction', password) -> None:
        pass

    @abstractmethod
    def get_pubkey_derivation(self, pubkey: bytes,
                              txinout: Union['PartialTxInput', 'PartialTxOutput'],
                              *, only_der_suffix=True) \
            -> Union[Sequence[int], str, None]:
        """Returns either a derivation int-list if the pubkey can be HD derived from this keystore,
        the pubkey itself (hex) if the pubkey belongs to the keystore but not HD derived,
        or None if the pubkey is unrelated.
        """
        pass

    @abstractmethod
    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
        pass

    def find_my_pubkey_in_txinout(
            self, txinout: Union['PartialTxInput', 'PartialTxOutput'],
            *, only_der_suffix: bool = False
    ) -> Tuple[Optional[bytes], Optional[List[int]]]:
        # note: we assume that this cosigner only has one pubkey in this txin/txout
        for pubkey in txinout.bip32_paths:
            path = self.get_pubkey_derivation(pubkey, txinout, only_der_suffix=only_der_suffix)
            if path and not isinstance(path, (str, bytes)):
                return pubkey, list(path)
        return None, None

    def can_have_deterministic_lightning_xprv(self) -> bool:
        return False

    def has_support_for_slip_19_ownership_proofs(self) -> bool:
        return False

    def add_slip_19_ownership_proofs_to_tx(self, tx: 'PartialTransaction', *, password) -> None:
        raise NotImplementedError()


class Software_KeyStore(KeyStore):

    def __init__(self, d: dict):
        KeyStore.__init__(self)
        self.pw_hash_version = d.get('pw_hash_version', 1)
        if self.pw_hash_version not in SUPPORTED_PW_HASH_VERSIONS:
            raise UnsupportedPasswordHashVersion(self.pw_hash_version)

    def may_have_password(self):
        return not self.is_watching_only()

    def sign_message(self, sequence, message, password, *, script_type=None) -> bytes:
        privkey, compressed = self.get_private_key(sequence, password)
        key = ecc.ECPrivkey(privkey)
        return bitcoin.ecdsa_sign_usermessage(key, message, is_compressed=compressed)

    def decrypt_message(self, sequence, message, password) -> bytes:
        privkey, compressed = self.get_private_key(sequence, password)
        ec = ecc.ECPrivkey(privkey)
        decrypted = crypto.ecies_decrypt_message(ec, message)
        return decrypted

    def sign_transaction(self, tx, password):
        if self.is_watching_only():
            return
        # Raise if password is not correct.
        self.check_password(password)
        # Add private keys
        keypairs = {}
        pubkey_to_deriv_map = self._get_tx_derivations(tx)
        for pubkey, deriv in pubkey_to_deriv_map.items():
            privkey, is_compressed = self.get_private_key(deriv, password)
            keypairs[pubkey] = privkey
        # Sign
        if keypairs:
            tx.sign(keypairs)

    @abstractmethod
    def update_password(self, old_password, new_password) -> None:
        pass

    @abstractmethod
    def check_password(self, password: Optional[str]) -> None:
        """Raises InvalidPassword if password is not correct"""
        pass

    @abstractmethod
    def get_private_key(self, sequence: 'AddressIndexGeneric', password) -> Tuple[bytes, bool]:
        """Returns (privkey, is_compressed)"""
        pass


class Imported_KeyStore(Software_KeyStore):
    # keystore for imported private keys

    type = 'imported'

    def __init__(self, d: dict):
        Software_KeyStore.__init__(self, d)
        self.keypairs = d.get('keypairs', {})  # type: Dict[str, str]

    def is_deterministic(self):
        return False

    def dump(self):
        return {
            'type': self.type,
            'keypairs': self.keypairs,
            'pw_hash_version': self.pw_hash_version,
        }

    def can_import(self):
        return True

    @also_test_none_password
    def check_password(self, password):
        pubkey = list(self.keypairs.keys())[0]
        self.get_private_key(pubkey, password)

    def import_privkey(self, sec: str, password) -> Tuple[str, str]:
        txin_type, privkey, compressed = deserialize_privkey(sec)
        pubkey = ecc.ECPrivkey(privkey).get_public_key_hex(compressed=compressed)
        # re-serialize the key so the internal storage format is consistent
        serialized_privkey = serialize_privkey(
            privkey, compressed, txin_type, internal_use=True)
        # NOTE: if the same pubkey is reused for multiple addresses (script types),
        # there will only be one pubkey-privkey pair for it in self.keypairs,
        # and the privkey will encode a txin_type but that txin_type cannot be trusted.
        # Removing keys complicates this further.
        self.keypairs[pubkey] = pw_encode(serialized_privkey, password, version=self.pw_hash_version)
        return txin_type, pubkey

    def import_private_keys(self, keys: Sequence[str], password: Optional[str]):
        good_inputs = []  # type: List[Tuple[str, bytes]]
        bad_keys = []  # type: List[Tuple[str, str]]
        for key in keys:
            try:
                txin_type, pubkey = self.import_privkey(key, password)
            except Exception as e:
                bad_keys.append((key, 'invalid private key' + f': {e}'))
                continue
            if txin_type not in ('p2pkh', 'p2wpkh', 'p2wpkh-p2sh'):
                bad_keys.append((key, 'not implemented type' + f': {txin_type}'))
                continue
            good_inputs.append((txin_type, pubkey))
        return good_inputs, bad_keys

    def delete_imported_key(self, key: str) -> None:
        self.keypairs.pop(key)

    def get_private_key(self, pubkey: str, password):
        sec = pw_decode(self.keypairs[pubkey], password, version=self.pw_hash_version)
        try:
            txin_type, privkey, compressed = deserialize_privkey(sec)
        except BaseDecodeError as e:
            raise InvalidPassword() from e
        if pubkey != ecc.ECPrivkey(privkey).get_public_key_hex(compressed=compressed):
            raise InvalidPassword()
        return privkey, compressed

    def get_pubkey_derivation(self, pubkey, txin, *, only_der_suffix=True):
        if pubkey.hex() in self.keypairs:
            return pubkey.hex()
        return None

    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
        if sequence in self.keypairs:
            return PubkeyProvider(
                origin=None,
                pubkey=sequence,
                deriv_path=None,
            )
        return None

    def update_password(self, old_password, new_password):
        self.check_password(old_password)
        if new_password == '':
            new_password = None
        for k, v in self.keypairs.items():
            b = pw_decode(v, old_password, version=self.pw_hash_version)
            c = pw_encode(b, new_password, version=PW_HASH_VERSION_LATEST)
            self.keypairs[k] = c
        self.pw_hash_version = PW_HASH_VERSION_LATEST


class Deterministic_KeyStore(Software_KeyStore):

    def __init__(self, d: dict):
        Software_KeyStore.__init__(self, d)
        self.seed = d.get('seed', '')  # only electrum seeds
        self.passphrase = d.get('passphrase', '')
        self._seed_type = d.get('seed_type', None)  # only electrum seeds

    def is_deterministic(self):
        return True

    def dump(self):
        d = {
            'type': self.type,
            'pw_hash_version': self.pw_hash_version,
        }
        if self.seed:
            d['seed'] = self.seed
        if self.passphrase:
            d['passphrase'] = self.passphrase
        if self._seed_type:
            d['seed_type'] = self._seed_type
        return d

    def has_seed(self):
        return bool(self.seed)

    def get_seed_type(self) -> Optional[str]:
        return self._seed_type

    def is_watching_only(self):
        return not self.has_seed()

    @abstractmethod
    def format_seed(self, seed: str) -> str:
        pass

    def add_seed(self, seed: str) -> None:
        if self.seed:
            raise Exception("a seed exists")
        self.seed = self.format_seed(seed)
        self._seed_type = calc_seed_type(seed) or None

    def get_seed(self, password) -> str:
        if not self.has_seed():
            raise Exception("This wallet has no seed words")
        return pw_decode(self.seed, password, version=self.pw_hash_version)

    def get_passphrase(self, password) -> str:
        if self.passphrase:
            return pw_decode(self.passphrase, password, version=self.pw_hash_version)
        else:
            return ''


class MasterPublicKeyMixin(ABC):

    def __init__(self):
        self._pubkey_cache = LRUCache(maxsize=10**4)  # type: LRUCache[Sequence[int], bytes]  # path->pubkey

    @abstractmethod
    def get_master_public_key(self) -> str:
        pass

    @abstractmethod
    def get_derivation_prefix(self) -> Optional[str]:
        """Returns to bip32 path from some root node to self.xpub
        Note that the return value might be None; if it is unknown.
        """
        pass

    @abstractmethod
    def get_root_fingerprint(self) -> Optional[str]:
        """Returns the bip32 fingerprint of the top level node.
        This top level node is the node at the beginning of the derivation prefix,
        i.e. applying the derivation prefix to it will result self.xpub
        Note that the return value might be None; if it is unknown.
        """
        pass

    @abstractmethod
    def get_fp_and_derivation_to_be_used_in_partial_tx(
            self,
            der_suffix: Sequence[int],
            *,
            only_der_suffix: bool,
    ) -> Tuple[bytes, Sequence[int]]:
        """Returns fingerprint and derivation path corresponding to a derivation suffix.
        The fingerprint is either the root fp or the intermediate fp, depending on what is available
        and 'only_der_suffix', and the derivation path is adjusted accordingly.
        """
        pass

    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
        return None

    def derive_pubkey(self, for_change: int, n: int) -> bytes:
        key = (for_change, n)
        if key not in self._pubkey_cache:
            self._pubkey_cache[key] = self._derive_pubkey(*key)
        return self._pubkey_cache[key]

    @abstractmethod
    def _derive_pubkey(self, for_change: int, n: int) -> bytes:
        """Returns pubkey at given path.
        May raise CannotDerivePubkey.
        """
        pass

    def get_pubkey_derivation(
            self,
            pubkey: bytes,
            txinout: Union['PartialTxInput', 'PartialTxOutput'],
            *,
            only_der_suffix=True,
    ) -> Union[Sequence[int], str, None]:
        EXPECTED_DER_SUFFIX_LEN = 2
        def test_der_suffix_against_pubkey(der_suffix: Sequence[int], pubkey: bytes) -> bool:
            if len(der_suffix) != EXPECTED_DER_SUFFIX_LEN:
                return False
            try:
                if pubkey != self.derive_pubkey(*der_suffix):
                    return False
            except CannotDerivePubkey:
                return False
            return True

        if pubkey not in txinout.bip32_paths:
            return None
        fp_found, path_found = txinout.bip32_paths[pubkey]
        der_suffix = None
        full_path = None
        # 1. try fp against our root
        ks_root_fingerprint_hex = self.get_root_fingerprint()
        ks_der_prefix_str = self.get_derivation_prefix()
        ks_der_prefix = convert_bip32_strpath_to_intpath(ks_der_prefix_str) if ks_der_prefix_str else None
        if (ks_root_fingerprint_hex is not None and ks_der_prefix is not None and
                fp_found.hex() == ks_root_fingerprint_hex):
            if path_found[:len(ks_der_prefix)] == ks_der_prefix:
                der_suffix = path_found[len(ks_der_prefix):]
                if not test_der_suffix_against_pubkey(der_suffix, pubkey):
                    der_suffix = None
        # 2. try fp against our intermediate fingerprint
        if (der_suffix is None and isinstance(self, Xpub) and
                fp_found == self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()):
            der_suffix = path_found
            if not test_der_suffix_against_pubkey(der_suffix, pubkey):
                der_suffix = None
        # 3. hack/bruteforce: ignore fp and check pubkey anyway
        #    This is only to resolve the following scenario/problem:
        #    problem: if we don't know our root fp, but tx contains root fp and full path,
        #             we will miss the pubkey (false negative match). Though it might still work
        #             within gap limit due to tx.add_info_from_wallet overwriting the fields.
        #             Example: keystore has intermediate xprv without root fp; tx contains root fp and full path.
        if der_suffix is None:
            der_suffix = path_found[-EXPECTED_DER_SUFFIX_LEN:]
            if not test_der_suffix_against_pubkey(der_suffix, pubkey):
                der_suffix = None
        # if all attempts/methods failed, we give up now:
        if der_suffix is None:
            return None
        if ks_der_prefix is not None:
            full_path = ks_der_prefix + list(der_suffix)
        return der_suffix if only_der_suffix else full_path


class Xpub(MasterPublicKeyMixin):

    def __init__(self, *, derivation_prefix: str = None, root_fingerprint: str = None):
        MasterPublicKeyMixin.__init__(self)
        self.xpub = None
        self.xpub_receive = None
        self.xpub_change = None
        self._xpub_bip32_node = None  # type: Optional[BIP32Node]

        # "key origin" info (subclass should persist these):
        self._derivation_prefix = derivation_prefix  # type: Optional[str]
        self._root_fingerprint = root_fingerprint  # type: Optional[str]

    def get_master_public_key(self):
        return self.xpub

    def get_bip32_node_for_xpub(self) -> Optional[BIP32Node]:
        if self._xpub_bip32_node is None:
            if self.xpub is None:
                return None
            self._xpub_bip32_node = BIP32Node.from_xkey(self.xpub)
        return self._xpub_bip32_node

    def get_derivation_prefix(self) -> Optional[str]:
        if self._derivation_prefix is None:
            return None
        return normalize_bip32_derivation(self._derivation_prefix)

    def get_root_fingerprint(self) -> Optional[str]:
        return self._root_fingerprint

    def get_fp_and_derivation_to_be_used_in_partial_tx(
            self,
            der_suffix: Sequence[int],
            *,
            only_der_suffix: bool,
    ) -> Tuple[bytes, Sequence[int]]:
        fingerprint_hex = self.get_root_fingerprint()
        der_prefix_str = self.get_derivation_prefix()
        if not only_der_suffix and fingerprint_hex is not None and der_prefix_str is not None:
            # use root fp, and true full path
            fingerprint_bytes = bfh(fingerprint_hex)
            der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
        else:
            # use intermediate fp, and claim der suffix is the full path
            fingerprint_bytes = self.get_bip32_node_for_xpub().calc_fingerprint_of_this_node()
            der_prefix_ints = convert_bip32_strpath_to_intpath('m')
        der_full = der_prefix_ints + list(der_suffix)
        return fingerprint_bytes, der_full

    def get_xpub_to_be_used_in_partial_tx(self, *, only_der_suffix: bool) -> str:
        assert self.xpub
        fp_bytes, der_full = self.get_fp_and_derivation_to_be_used_in_partial_tx(der_suffix=[],
                                                                                 only_der_suffix=only_der_suffix)
        bip32node = self.get_bip32_node_for_xpub()
        depth = len(der_full)
        child_number_int = der_full[-1] if len(der_full) >= 1 else 0
        child_number_bytes = child_number_int.to_bytes(length=4, byteorder="big")
        fingerprint = bytes(4) if depth == 0 else bip32node.fingerprint
        bip32node = bip32node._replace(
            depth=depth,
            fingerprint=fingerprint,
            child_number=child_number_bytes,
            # only put plain xpubs (not ypub/zpub) in PSBTs:
            xtype="standard",
        )
        return bip32node.to_xpub()

    def get_key_origin_info(self) -> Optional[KeyOriginInfo]:
        fp_bytes, der_full = self.get_fp_and_derivation_to_be_used_in_partial_tx(
            der_suffix=[], only_der_suffix=False)
        origin = KeyOriginInfo(fingerprint=fp_bytes, path=der_full)
        return origin

    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
        strpath = convert_bip32_intpath_to_strpath(sequence)
        strpath = strpath[1:]  # cut leading "m"
        bip32node = self.get_bip32_node_for_xpub()
        return PubkeyProvider(
            origin=self.get_key_origin_info(),
            pubkey=bip32node._replace(xtype="standard").to_xkey(),
            deriv_path=strpath,
        )

    def add_key_origin_from_root_node(self, *, derivation_prefix: str, root_node: BIP32Node) -> None:
        assert self.xpub
        # try to derive ourselves from what we were given
        child_node1 = root_node.subkey_at_private_derivation(derivation_prefix)
        child_pubkey_bytes1 = child_node1.eckey.get_public_key_bytes(compressed=True)
        child_node2 = self.get_bip32_node_for_xpub()
        child_pubkey_bytes2 = child_node2.eckey.get_public_key_bytes(compressed=True)
        if child_pubkey_bytes1 != child_pubkey_bytes2:
            raise Exception("(xpub, derivation_prefix, root_node) inconsistency")
        self.add_key_origin(derivation_prefix=derivation_prefix,
                            root_fingerprint=root_node.calc_fingerprint_of_this_node().hex().lower())

    def add_key_origin(self, *, derivation_prefix: str = None, root_fingerprint: str = None) -> None:
        assert self.xpub
        if not (root_fingerprint is None or (is_hex_str(root_fingerprint) and len(root_fingerprint) == 8)):
            raise Exception("root fp must be 8 hex characters")
        derivation_prefix = normalize_bip32_derivation(derivation_prefix)
        if not is_xkey_consistent_with_key_origin_info(self.xpub,
                                                       derivation_prefix=derivation_prefix,
                                                       root_fingerprint=root_fingerprint):
            raise Exception("xpub inconsistent with provided key origin info")
        if root_fingerprint is not None:
            self._root_fingerprint = root_fingerprint
        if derivation_prefix is not None:
            self._derivation_prefix = derivation_prefix
        self.is_requesting_to_be_rewritten_to_wallet_file = True

    def _derive_pubkey(self, for_change: int, n: int) -> bytes:
        for_change = int(for_change)
        if for_change not in (0, 1):
            raise CannotDerivePubkey("forbidden path")
        xpub = self.xpub_change if for_change else self.xpub_receive
        if xpub is None:
            rootnode = self.get_bip32_node_for_xpub()
            xpub = rootnode.subkey_at_public_derivation((for_change,)).to_xpub()
            if for_change:
                self.xpub_change = xpub
            else:
                self.xpub_receive = xpub
        return self.get_pubkey_from_xpub(xpub, (n,))

    @classmethod
    def get_pubkey_from_xpub(cls, xpub: str, sequence) -> bytes:
        node = BIP32Node.from_xkey(xpub).subkey_at_public_derivation(sequence)
        return node.eckey.get_public_key_bytes(compressed=True)


class BIP32_KeyStore(Xpub, Deterministic_KeyStore):

    type = 'bip32'

    def __init__(self, d: dict):
        Xpub.__init__(self, derivation_prefix=d.get('derivation'), root_fingerprint=d.get('root_fingerprint'))
        Deterministic_KeyStore.__init__(self, d)
        self.xpub = d.get('xpub')
        self.xprv = d.get('xprv')

    def watching_only_keystore(self):
        return BIP32_KeyStore({
            'xpub': self.xpub,
            'root_fingerprint': self.get_root_fingerprint(),
            'derivation': self.get_derivation_prefix(),
        })

    def format_seed(self, seed):
        return ' '.join(seed.split())

    def dump(self):
        d = Deterministic_KeyStore.dump(self)
        d['xpub'] = self.xpub
        d['xprv'] = self.xprv
        d['derivation'] = self.get_derivation_prefix()
        d['root_fingerprint'] = self.get_root_fingerprint()
        return d

    def get_master_private_key(self, password) -> str:
        return pw_decode(self.xprv, password, version=self.pw_hash_version)

    @also_test_none_password
    def check_password(self, password):
        xprv = pw_decode(self.xprv, password, version=self.pw_hash_version)
        try:
            bip32node = BIP32Node.from_xkey(xprv)
        except BaseDecodeError as e:
            raise InvalidPassword() from e
        if bip32node.chaincode != self.get_bip32_node_for_xpub().chaincode:
            raise InvalidPassword()

    def update_password(self, old_password, new_password):
        self.check_password(old_password)
        if new_password == '':
            new_password = None
        if self.has_seed():
            decoded = self.get_seed(old_password)
            self.seed = pw_encode(decoded, new_password, version=PW_HASH_VERSION_LATEST)
        if self.passphrase:
            decoded = self.get_passphrase(old_password)
            self.passphrase = pw_encode(decoded, new_password, version=PW_HASH_VERSION_LATEST)
        if self.xprv is not None:
            b = pw_decode(self.xprv, old_password, version=self.pw_hash_version)
            self.xprv = pw_encode(b, new_password, version=PW_HASH_VERSION_LATEST)
        self.pw_hash_version = PW_HASH_VERSION_LATEST

    def is_watching_only(self):
        return self.xprv is None

    def add_xpub(self, xpub: str) -> None:
        assert is_xpub(xpub)
        self.xpub = xpub
        root_fingerprint, derivation_prefix = bip32.root_fp_and_der_prefix_from_xkey(xpub)
        self.add_key_origin(derivation_prefix=derivation_prefix, root_fingerprint=root_fingerprint)

    def add_xprv(self, xprv: str) -> None:
        assert is_xprv(xprv)
        self.xprv = xprv
        self.add_xpub(bip32.xpub_from_xprv(xprv))

    def add_xprv_from_seed(self, bip32_seed: bytes, *, xtype: str, derivation: str) -> None:
        rootnode = BIP32Node.from_rootseed(bip32_seed, xtype=xtype)
        node = rootnode.subkey_at_private_derivation(derivation)
        self.add_xprv(node.to_xprv())
        self.add_key_origin_from_root_node(derivation_prefix=derivation, root_node=rootnode)

    def get_private_key(self, sequence: Sequence[int], password):
        xprv = self.get_master_private_key(password)
        node = BIP32Node.from_xkey(xprv).subkey_at_private_derivation(sequence)
        pk = node.eckey.get_secret_bytes()
        return pk, True

    def can_have_deterministic_lightning_xprv(self):
        if (self.get_seed_type() == 'segwit'
                and self.get_bip32_node_for_xpub().xtype == 'p2wpkh'):
            return True
        return False

    def get_lightning_xprv(self, password) -> str:
        assert self.can_have_deterministic_lightning_xprv()
        xprv = self.get_master_private_key(password)
        rootnode = BIP32Node.from_xkey(xprv)
        node = rootnode.subkey_at_private_derivation("m/67'/")
        return node.to_xprv()

class Old_KeyStore(MasterPublicKeyMixin, Deterministic_KeyStore):

    type = 'old'

    def __init__(self, d: dict):
        MasterPublicKeyMixin.__init__(self)
        Deterministic_KeyStore.__init__(self, d)
        self.mpk = d.get('mpk')  # type: Optional[str]
        self._root_fingerprint = None

    def watching_only_keystore(self):
        return Old_KeyStore({'mpk': self.mpk})

    def _get_hex_seed(self, password) -> str:
        if not is_hex_str(self.seed) and password is None:
            raise InvalidPassword()
        hex_str = pw_decode(self.seed, password, version=self.pw_hash_version)
        assert is_hex_str(hex_str), f"expected hex str, got {type(hex_str)} with {len(hex_str)=}"
        return hex_str

    def dump(self):
        d = Deterministic_KeyStore.dump(self)
        d['mpk'] = self.mpk
        return d

    def add_seed(self, seed):
        Deterministic_KeyStore.add_seed(self, seed)
        hex_seed = self._get_hex_seed(None)
        self.mpk = self.mpk_from_seed(hex_seed)

    def add_master_public_key(self, mpk: str) -> None:
        self.mpk = mpk

    def format_seed(self, seed):
        """Returns seed in hex format.

        seed: either in hex or as mnemonic words
        """
        from . import old_mnemonic, mnemonic
        seed = mnemonic.normalize_text(seed)
        # see if seed was entered as hex
        if seed:
            try:
                bfh(seed)
                return str(seed)
            except Exception:
                pass
        words = seed.split()
        seed = old_mnemonic.mn_decode(words)
        if not seed:
            raise Exception("Invalid seed")
        return seed

    def get_seed(self, password):
        from . import old_mnemonic
        hex_seed = self._get_hex_seed(password)
        return ' '.join(old_mnemonic.mn_encode(hex_seed))

    @classmethod
    def mpk_from_seed(cls, hex_seed: str) -> str:
        secexp = cls.stretch_key(hex_seed)
        privkey = ecc.ECPrivkey.from_secret_scalar(secexp)
        return privkey.get_public_key_hex(compressed=False)[2:]

    @classmethod
    def stretch_key(cls, hex_seed: str) -> int:
        assert is_hex_str(hex_seed), f"expected hex str, got {type(hex_seed)} with {len(hex_seed)=}"
        encoded_hex_seed = hex_seed.encode('ascii')
        x = encoded_hex_seed
        for i in range(100000):
            x = hashlib.sha256(x + encoded_hex_seed).digest()
        return string_to_number(x)

    @classmethod
    def get_sequence(cls, mpk: str, for_change: int, n: int) -> int:
        return string_to_number(sha256d(("%d:%d:"%(n, for_change)).encode('ascii') + bfh(mpk)))

    @classmethod
    def get_pubkey_from_mpk(cls, mpk: str, for_change: int, n: int) -> bytes:
        z = cls.get_sequence(mpk, for_change, n)
        master_public_key = ecc.ECPubkey(bfh('04'+mpk))
        public_key = master_public_key + z*ecc.GENERATOR
        return public_key.get_public_key_bytes(compressed=False)

    def _derive_pubkey(self, for_change, n) -> bytes:
        for_change = int(for_change)
        if for_change not in (0, 1):
            raise CannotDerivePubkey("forbidden path")
        return self.get_pubkey_from_mpk(self.mpk, for_change, n)

    def _get_private_key_from_stretched_exponent(self, for_change: int, n: int, secexp: int) -> bytes:
        secexp = (secexp + self.get_sequence(self.mpk, for_change, n)) % ecc.CURVE_ORDER
        pk = int.to_bytes(secexp, length=32, byteorder='big', signed=False)
        return pk

    def get_private_key(self, sequence: Sequence[int], password):
        hex_seed = self._get_hex_seed(password)
        secexp = self.stretch_key(hex_seed)
        self._check_seed(hex_seed, secexp=secexp)
        for_change, n = sequence
        assert isinstance(for_change, int), type(for_change)
        assert isinstance(n, int), type(n)
        pk = self._get_private_key_from_stretched_exponent(for_change, n, secexp)
        return pk, False

    def _check_seed(self, hex_seed: str, *, secexp: int = None) -> None:
        if secexp is None:
            secexp = self.stretch_key(hex_seed)
        master_private_key = ecc.ECPrivkey.from_secret_scalar(secexp)
        master_public_key = master_private_key.get_public_key_bytes(compressed=False)[1:]
        if master_public_key != bfh(self.mpk):
            raise InvalidPassword()

    @also_test_none_password
    def check_password(self, password):
        hex_seed = self._get_hex_seed(password)
        self._check_seed(hex_seed)

    def get_master_public_key(self):
        return self.mpk

    def get_derivation_prefix(self) -> str:
        return 'm'

    def get_root_fingerprint(self) -> str:
        if self._root_fingerprint is None:
            master_public_key = ecc.ECPubkey(bfh('04'+self.mpk))
            xfp = hash_160(master_public_key.get_public_key_bytes(compressed=True))[0:4]
            self._root_fingerprint = xfp.hex().lower()
        return self._root_fingerprint

    def get_fp_and_derivation_to_be_used_in_partial_tx(
            self,
            der_suffix: Sequence[int],
            *,
            only_der_suffix: bool,
    ) -> Tuple[bytes, Sequence[int]]:
        fingerprint_hex = self.get_root_fingerprint()
        der_prefix_str = self.get_derivation_prefix()
        fingerprint_bytes = bfh(fingerprint_hex)
        der_prefix_ints = convert_bip32_strpath_to_intpath(der_prefix_str)
        der_full = der_prefix_ints + list(der_suffix)
        return fingerprint_bytes, der_full

    def get_pubkey_provider(self, sequence: 'AddressIndexGeneric') -> Optional[PubkeyProvider]:
        return PubkeyProvider(
            origin=None,
            pubkey=self.derive_pubkey(*sequence).hex(),
            deriv_path=None,
        )

    def update_password(self, old_password, new_password):
        self.check_password(old_password)
        if new_password == '':
            new_password = None
        if self.has_seed():
            decoded = pw_decode(self.seed, old_password, version=self.pw_hash_version)
            self.seed = pw_encode(decoded, new_password, version=PW_HASH_VERSION_LATEST)
        self.pw_hash_version = PW_HASH_VERSION_LATEST


class Hardware_KeyStore(Xpub, KeyStore):
    hw_type: str
    device: str
    plugin: 'HW_PluginBase'
    thread: Optional['TaskThread'] = None

    type = 'hardware'

    def __init__(self, d):
        Xpub.__init__(self, derivation_prefix=d.get('derivation'), root_fingerprint=d.get('root_fingerprint'))
        KeyStore.__init__(self)
        # Errors and other user interaction is done through the wallet's
        # handler.  The handler is per-window and preserved across
        # device reconnects
        self.xpub = d.get('xpub')
        self.label = d.get('label')  # type: Optional[str]
        self.soft_device_id = d.get('soft_device_id')  # type: Optional[str]
        self.handler = None  # type: Optional[HardwareHandlerBase]
        run_hook('init_keystore', self)

    def watching_only_keystore(self):
        return BIP32_KeyStore({
            'xpub': self.xpub,
            'root_fingerprint': self.get_root_fingerprint(),
            'derivation': self.get_derivation_prefix(),
        })

    def set_label(self, label: Optional[str]) -> None:
        self.label = label

    def may_have_password(self):
        return False

    def is_deterministic(self):
        return True

    def get_type_text(self) -> str:
        return f'hw[{self.hw_type}]'

    def dump(self):
        return {
            'type': self.type,
            'hw_type': self.hw_type,
            'xpub': self.xpub,
            'derivation': self.get_derivation_prefix(),
            'root_fingerprint': self.get_root_fingerprint(),
            'label': self.label,
            'soft_device_id': self.soft_device_id,
        }

    def is_watching_only(self):
        """The wallet is not watching-only; the user will be prompted for
        pin and passphrase as appropriate when needed."""
        assert not self.has_seed()
        return False

    def get_client(
            self,
            force_pair: bool = True,
            *,
            devices: Sequence['Device'] = None,
            allow_user_interaction: bool = True,
    ) -> Optional['HardwareClientBase']:
        return self.plugin.get_client(
            self,
            force_pair=force_pair,
            devices=devices,
            allow_user_interaction=allow_user_interaction,
        )

    def get_password_for_storage_encryption(self) -> str:
        client = self.get_client()
        return client.get_password_for_storage_encryption()

    def has_usable_connection_with_device(self) -> bool:
        # we try to create a client even if there isn't one already,
        # but do not prompt the user if auto-select fails:
        client = self.get_client(
            force_pair=True,
            allow_user_interaction=False,
        )
        if client is None:
            return False
        return client.has_usable_connection_with_device()

    def ready_to_sign(self):
        return super().ready_to_sign() and self.has_usable_connection_with_device()

    def opportunistically_fill_in_missing_info_from_device(self, client: 'HardwareClientBase'):
        assert client is not None
        if self._root_fingerprint is None:
            self._root_fingerprint = client.request_root_fingerprint_from_device()
            self.is_requesting_to_be_rewritten_to_wallet_file = True
        if self.label != client.label():
            self.label = client.label()
            self.is_requesting_to_be_rewritten_to_wallet_file = True
        if self.soft_device_id != client.get_soft_device_id():
            self.soft_device_id = client.get_soft_device_id()
            self.is_requesting_to_be_rewritten_to_wallet_file = True

    def pairing_code(self) -> Optional[str]:
        """Used by the DeviceMgr to keep track of paired hw devices."""
        if not self.soft_device_id:
            return None
        return f"{self.plugin.name}/{self.soft_device_id}"


KeyStoreWithMPK = Union[KeyStore, MasterPublicKeyMixin]  # intersection really...
AddressIndexGeneric = Union[Sequence[int], str]  # can be hex pubkey str


def bip39_normalize_passphrase(passphrase: str):
    return normalize('NFKD', passphrase or '')


def bip39_to_seed(mnemonic: str, *, passphrase: Optional[str]) -> bytes:
    import hashlib
    passphrase = passphrase or ""
    PBKDF2_ROUNDS = 2048
    mnemonic = normalize('NFKD', ' '.join(mnemonic.split()))
    passphrase = bip39_normalize_passphrase(passphrase)
    return hashlib.pbkdf2_hmac('sha512', mnemonic.encode('utf-8'),
        b'mnemonic' + passphrase.encode('utf-8'), iterations = PBKDF2_ROUNDS)


def bip39_is_checksum_valid(
        mnemonic: str,
        *,
        wordlist: Wordlist = None,
) -> Tuple[bool, bool]:
    """Test checksum of bip39 mnemonic assuming English wordlist.
    Returns tuple (is_checksum_valid, is_wordlist_valid)
    """
    words = [normalize('NFKD', word) for word in mnemonic.split()]
    words_len = len(words)
    if wordlist is None:
        wordlist = Wordlist.from_file("english.txt")
    n = len(wordlist)
    i = 0
    words.reverse()
    while words:
        w = words.pop()
        try:
            k = wordlist.index(w)
        except ValueError:
            return False, False
        i = i*n + k
    if words_len not in [12, 15, 18, 21, 24]:
        return False, True
    checksum_length = 11 * words_len // 33  # num bits
    entropy_length = 32 * checksum_length  # num bits
    entropy = i >> checksum_length
    checksum = i % 2**checksum_length
    entropy_bytes = int.to_bytes(entropy, length=entropy_length//8, byteorder="big")
    hashed = int.from_bytes(sha256(entropy_bytes), byteorder="big")
    calculated_checksum = hashed >> (256 - checksum_length)
    return checksum == calculated_checksum, True


def from_bip43_rootseed(
    root_seed: bytes,
    *,
    derivation: str,
    xtype: Optional[str] = None,
):
    k = BIP32_KeyStore({})
    if xtype is None:
        xtype = xtype_from_derivation(derivation)
    k.add_xprv_from_seed(root_seed, xtype=xtype, derivation=derivation)
    return k


PURPOSE48_SCRIPT_TYPES = {
    'p2wsh-p2sh': 1,  # specifically multisig
    'p2wsh': 2,       # specifically multisig
}
PURPOSE48_SCRIPT_TYPES_INV = inv_dict(PURPOSE48_SCRIPT_TYPES)


def xtype_from_derivation(derivation: str) -> str:
    """Returns the script type to be used for this derivation."""
    bip32_indices = convert_bip32_strpath_to_intpath(derivation)
    if len(bip32_indices) >= 1:
        if bip32_indices[0] == 84 + BIP32_PRIME:
            return 'p2wpkh'
        elif bip32_indices[0] == 49 + BIP32_PRIME:
            return 'p2wpkh-p2sh'
        elif bip32_indices[0] == 44 + BIP32_PRIME:
            return 'standard'
        elif bip32_indices[0] == 45 + BIP32_PRIME:
            return 'standard'

    if len(bip32_indices) >= 4:
        if bip32_indices[0] == 48 + BIP32_PRIME:
            # m / purpose' / coin_type' / account' / script_type' / change / address_index
            script_type_int = bip32_indices[3] - BIP32_PRIME
            script_type = PURPOSE48_SCRIPT_TYPES_INV.get(script_type_int)
            if script_type is not None:
                return script_type
    return 'standard'


hw_keystores = {}  # type: Dict[str, Type[Hardware_KeyStore]]

def register_keystore(hw_type: str, constructor: Type[Hardware_KeyStore]) -> None:
    hw_keystores[hw_type] = constructor

def hardware_keystore(d) -> Hardware_KeyStore:
    hw_type = d['hw_type']
    if hw_type in hw_keystores:
        constructor = hw_keystores[hw_type]
        return constructor(d)
    raise WalletFileException(f'unknown hardware type: {hw_type}. '
                              f'hw_keystores: {list(hw_keystores)}')

def load_keystore(db: 'WalletDB', name: str) -> KeyStore:
    # deepcopy object to avoid keeping a pointer to db.data
    # note: this is needed as type(wallet.db.get("keystore")) != StoredDict
    d = copy.deepcopy(db.get(name, {}))
    t = d.get('type')
    if not t:
        raise WalletFileException(
            'Wallet format requires update.\n'
            'Cannot find keystore for name {}'.format(name))
    keystore_constructors = {ks.type: ks for ks in [Old_KeyStore, Imported_KeyStore, BIP32_KeyStore]}
    keystore_constructors['hardware'] = hardware_keystore
    try:
        ks_constructor = keystore_constructors[t]
    except KeyError:
        raise WalletFileException(f'Unknown type {t} for keystore named {name}')
    k = ks_constructor(d)
    return k


def is_old_mpk(mpk: str) -> bool:
    try:
        int(mpk, 16)  # test if hex string
    except Exception:
        return False
    if len(mpk) != 128:
        return False
    try:
        ecc.ECPubkey(bfh('04' + mpk))
    except Exception:
        return False
    return True


def is_address_list(text: str) -> bool:
    parts = text.split()
    return bool(parts) and all(bitcoin.is_address(x) for x in parts)


def get_private_keys(text: str, *, allow_spaces_inside_key=True, raise_on_error=False) -> Sequence[str]:
    if allow_spaces_inside_key:  # see #1612
        parts = text.split('\n')
        parts = map(lambda x: ''.join(x.split()), parts)
        parts = list(filter(bool, parts))
    else:
        parts = text.split()
    if bool(parts) and all(bitcoin.is_private_key(x, raise_on_error=raise_on_error) for x in parts):
        return parts
    return []


def is_private_key_list(text: str, *, allow_spaces_inside_key: bool = True, raise_on_error: bool = False) -> bool:
    return bool(get_private_keys(text,
                                 allow_spaces_inside_key=allow_spaces_inside_key,
                                 raise_on_error=raise_on_error))


def is_master_key(x: str) -> bool:
    return is_old_mpk(x) or is_bip32_key(x)


def is_bip32_key(x: str) -> bool:
    return is_xprv(x) or is_xpub(x)


def bip44_derivation(account_id: int, bip43_purpose: int = 44) -> str:
    coin = constants.net.BIP44_COIN_TYPE
    der = "m/%d'/%d'/%d'" % (bip43_purpose, coin, int(account_id))
    return normalize_bip32_derivation(der)


def purpose48_derivation(account_id: int, xtype: str) -> str:
    # m / purpose' / coin_type' / account' / script_type' / change / address_index
    bip43_purpose = 48
    coin = constants.net.BIP44_COIN_TYPE
    account_id = int(account_id)
    script_type_int = PURPOSE48_SCRIPT_TYPES.get(xtype)
    if script_type_int is None:
        raise Exception('unknown xtype: {}'.format(xtype))
    der = "m/%d'/%d'/%d'/%d'" % (bip43_purpose, coin, account_id, script_type_int)
    return normalize_bip32_derivation(der)


def from_seed(seed: str, *, passphrase: Optional[str], for_multisig: bool = False) -> Union[BIP32_KeyStore, Old_KeyStore]:
    passphrase = passphrase or ""
    t = calc_seed_type(seed)
    if t == 'old':
        if passphrase:
            raise Exception("'old'-type electrum seed cannot have passphrase")
        keystore = Old_KeyStore({})
        keystore.add_seed(seed)
    elif t in ['standard', 'segwit']:
        keystore = BIP32_KeyStore({})
        keystore.add_seed(seed)
        keystore.passphrase = passphrase
        bip32_seed = Mnemonic.mnemonic_to_seed(seed, passphrase=passphrase)
        if t == 'standard':
            der = "m/"
            xtype = 'standard'
        else:
            der = "m/1'/" if for_multisig else "m/0'/"
            xtype = 'p2wsh' if for_multisig else 'p2wpkh'
        keystore.add_xprv_from_seed(bip32_seed, xtype=xtype, derivation=der)
    else:
        raise BitcoinException('Unexpected seed type {}'.format(repr(t)))
    return keystore

def from_private_key_list(text: str) -> Imported_KeyStore:
    keystore = Imported_KeyStore({})
    for x in get_private_keys(text):
        keystore.import_privkey(x, None)
    return keystore

def from_old_mpk(mpk: str) -> Old_KeyStore:
    keystore = Old_KeyStore({})
    keystore.add_master_public_key(mpk)
    return keystore

def from_xpub(xpub: str) -> BIP32_KeyStore:
    k = BIP32_KeyStore({})
    k.add_xpub(xpub)
    return k

def from_xprv(xprv: str) -> BIP32_KeyStore:
    k = BIP32_KeyStore({})
    k.add_xprv(xprv)
    return k

def from_master_key(text: str) -> Union[BIP32_KeyStore, Old_KeyStore]:
    if is_xprv(text):
        k = from_xprv(text)
    elif is_old_mpk(text):
        k = from_old_mpk(text)
    elif is_xpub(text):
        k = from_xpub(text)
    else:
        raise BitcoinException('Invalid master key')
    return k



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2025 The Electrum Developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from abc import abstractmethod, ABC
from typing import TYPE_CHECKING, Sequence, Optional, Type, Iterable, Any

from electrum.plugin import (BasePlugin, hook, Device, DeviceMgr,
                             assert_runs_in_hwd_thread, runs_in_hwd_thread)
from electrum.i18n import _
from electrum.bitcoin import is_address, opcodes
from electrum.util import versiontuple, UserFacingException, ChoiceItem
from electrum.transaction import TxOutput, PartialTransaction
from electrum.bip32 import BIP32Node
from electrum.storage import get_derivation_used_for_hw_device_encryption
from electrum.keystore import Xpub, Hardware_KeyStore

if TYPE_CHECKING:
    import threading
    from electrum.plugin import DeviceInfo
    from electrum.wallet import Abstract_Wallet
    from electrum.wizard import AbstractWizard


class HW_PluginBase(BasePlugin, ABC):
    keystore_class: Type['Hardware_KeyStore']
    libraries_available: bool
    SUPPORTED_XTYPES = ()

    # define supported library versions:  minimum_library <= x < maximum_library
    minimum_library = (0,)
    maximum_library = (float('inf'),)

    DEVICE_IDS: Iterable[Any]

    def __init__(self, parent, config, name):
        BasePlugin.__init__(self, parent, config, name)
        self.device = self.keystore_class.device
        self.keystore_class.plugin = self
        self._ignore_outdated_fw = False

    def is_enabled(self):
        return True

    def device_manager(self) -> 'DeviceMgr':
        return self.parent.device_manager

    def create_device_from_hid_enumeration(self, d: dict, *, product_key) -> Optional['Device']:
        # note: id_ needs to be unique between simultaneously connected devices,
        #       and ideally unchanged while a device is connected.
        # Older versions of hid don't provide interface_number
        interface_number = d.get('interface_number', -1)
        usage_page = d['usage_page']
        # id_=str(d['path']) in itself might be sufficient, but this had to be touched
        # a number of times already, so let's just go for the overkill approach:
        id_ = f"{d['path']},{d['serial_number']},{interface_number},{usage_page}"
        device = Device(path=d['path'],
                        interface_number=interface_number,
                        id_=id_,
                        product_key=product_key,
                        usage_page=usage_page,
                        transport_ui_string='hid')
        return device

    @hook
    def close_wallet(self, wallet: 'Abstract_Wallet'):
        for keystore in wallet.get_keystores():
            if isinstance(keystore, self.keystore_class):
                self.device_manager().unpair_pairing_code(keystore.pairing_code())
                if keystore.thread:
                    keystore.thread.stop()

    def get_client(self, keystore: 'Hardware_KeyStore', force_pair: bool = True, *,
                   devices: Sequence['Device'] = None,
                   allow_user_interaction: bool = True) -> Optional['HardwareClientBase']:
        devmgr = self.device_manager()
        handler = keystore.handler
        client = devmgr.client_for_keystore(self, handler, keystore, force_pair,
                                            devices=devices,
                                            allow_user_interaction=allow_user_interaction)
        return client

    def show_address(self, wallet: 'Abstract_Wallet', address, keystore: 'Hardware_KeyStore' = None):
        pass  # implemented in child classes

    def show_address_helper(self, wallet, address, keystore=None):
        if keystore is None:
            keystore = wallet.get_keystore()
        if not is_address(address):
            keystore.handler.show_error(_('Invalid Bitcoin Address'))
            return False
        if not wallet.is_mine(address):
            keystore.handler.show_error(_('Address not in wallet.'))
            return False
        if type(keystore) != self.keystore_class:
            return False
        return True

    def get_library_version(self) -> str:
        """Returns the version of the 3rd party python library
        for the hw wallet. For example '0.9.0'

        Returns 'unknown' if library is found but cannot determine version.
        Raises 'ImportError' if library is not found.
        Raises 'LibraryFoundButUnusable' if found but there was some problem (includes version num).
        """
        raise NotImplementedError()

    def check_libraries_available(self) -> bool:
        def version_str(t):
            return ".".join(str(i) for i in t)

        try:
            # this might raise ImportError or LibraryFoundButUnusable
            library_version = self.get_library_version()
            # if no exception so far, we might still raise LibraryFoundButUnusable
            if (library_version == 'unknown'
                    or versiontuple(library_version) < self.minimum_library
                    or versiontuple(library_version) >= self.maximum_library):
                raise LibraryFoundButUnusable(library_version=library_version)
        except ImportError as e:
            self.libraries_available_message = (
                _("Missing libraries for {}.").format(self.name)
                + f"\n    {e!r}"
            )
            return False
        except LibraryFoundButUnusable as e:
            library_version = e.library_version
            self.libraries_available_message = (
                    _("Library version for '{}' is incompatible.").format(self.name)
                    + '\nInstalled: {}, Needed: {} <= x < {}'
                    .format(library_version, version_str(self.minimum_library), version_str(self.maximum_library)))
            self.logger.warning(self.libraries_available_message)
            return False

        return True

    def get_library_not_available_message(self) -> str:
        if hasattr(self, 'libraries_available_message'):
            message = self.libraries_available_message
        else:
            message = _("Missing libraries for {}.").format(self.name)
        message += '\n' + _("Make sure you install it with python3")
        return message

    def set_ignore_outdated_fw(self):
        self._ignore_outdated_fw = True

    def is_outdated_fw_ignored(self) -> bool:
        return self._ignore_outdated_fw

    def create_client(self, device: 'Device',
                      handler: Optional['HardwareHandlerBase']) -> Optional['HardwareClientBase']:
        raise NotImplementedError()

    def create_handler(self, window) -> 'HardwareHandlerBase':
        # note: in Qt GUI, 'window' is either an ElectrumWindow or an QENewWalletWizard
        raise NotImplementedError()

    def can_recognize_device(self, device: Device) -> bool:
        """Whether the plugin thinks it can handle the given device.
        Used for filtering all connected hardware devices to only those by this vendor.
        """
        return device.product_key in self.DEVICE_IDS

    @abstractmethod
    def wizard_entry_for_device(self, device_info: 'DeviceInfo', *, new_wallet: bool) -> str:
        """Return view name for device
        """
        pass

    @hook
    def init_wallet_wizard(self, wizard: 'AbstractWizard') -> None:
        self.extend_wizard(wizard)

    @abstractmethod
    def extend_wizard(self, wizard: 'AbstractWizard') -> None:
        pass


class HardwareClientBase(ABC):
    handler = None  # type: Optional['HardwareHandlerBase']

    def __init__(self, *, plugin: 'HW_PluginBase'):
        assert_runs_in_hwd_thread()
        self.plugin = plugin

    def device_manager(self) -> 'DeviceMgr':
        return self.plugin.device_manager()

    @abstractmethod
    def is_pairable(self) -> bool:
        pass

    @abstractmethod
    def close(self):
        pass

    def timeout(self, cutoff) -> None:  # noqa: B027
        pass

    @abstractmethod
    def is_initialized(self) -> bool:
        """True if initialized, False if wiped."""
        pass

    def label(self) -> Optional[str]:
        """The name given by the user to the device.

        Note: labels are shown to the user to help distinguish their devices,
        and they are also used as a fallback to distinguish devices programmatically.
        So ideally, different devices would have different labels.
        """
        # When returning a constant here (i.e. not implementing the method in the way
        # it is supposed to work), make sure the return value is in electrum.plugin.PLACEHOLDER_HW_CLIENT_LABELS
        return " "

    def get_soft_device_id(self) -> Optional[str]:
        """An id-like string that is used to distinguish devices programmatically.
        This is a long term id for the device, that does not change between reconnects.
        This method should not prompt the user, i.e. no user interaction, as it is used
        during USB device enumeration (called for each unpaired device).
        Stored in the wallet file.
        """
        root_fp = self.request_root_fingerprint_from_device()
        return root_fp

    @abstractmethod
    def has_usable_connection_with_device(self) -> bool:
        pass

    @abstractmethod
    def get_xpub(self, bip32_path: str, xtype) -> str:
        pass

    @runs_in_hwd_thread
    def request_root_fingerprint_from_device(self) -> str:
        # digitalbitbox (at least) does not reveal xpubs corresponding to unhardened paths
        # so ask for a direct child, and read out fingerprint from that:
        child_of_root_xpub = self.get_xpub("m/0'", xtype='standard')
        root_fingerprint = BIP32Node.from_xkey(child_of_root_xpub).fingerprint.hex().lower()
        return root_fingerprint

    @runs_in_hwd_thread
    def get_password_for_storage_encryption(self) -> str:
        # note: using a different password based on hw device type is highly undesirable! see #5993
        derivation = get_derivation_used_for_hw_device_encryption()
        xpub = self.get_xpub(derivation, "standard")
        password = Xpub.get_pubkey_from_xpub(xpub, ()).hex()
        return password

    def device_model_name(self) -> Optional[str]:
        """Return the name of the model of this device, which might be displayed in the UI.
        E.g. for Trezor, "Trezor One" or "Trezor T".
        If this method is not defined for a plugin, the plugin name is used as default
        """
        return self.plugin.name


class HardwareClientDummy(HardwareClientBase):
    """Hw device we recognize but do not support.
    E.g. for Ledger HW.1 devices that we used to support in the past, but no longer do.
    This allows showing an error message to the user.
    """
    def __init__(self, *, plugin: 'HW_PluginBase', error_text: str):
        HardwareClientBase.__init__(self, plugin=plugin)
        self.error_text = error_text

    def get_xpub(self, bip32_path: str, xtype) -> str:
        raise Exception(self.error_text)

    def is_pairable(self) -> bool:
        return False

    def close(self):
        pass

    def is_initialized(self) -> bool:
        """True if initialized, False if wiped."""
        return True

    def label(self) -> Optional[str]:
        return "dummy_client"

    def has_usable_connection_with_device(self) -> bool:
        return True


class HardwareHandlerBase:
    """An interface between the GUI and the device handling logic for handling I/O."""
    win = None
    device: str

    def get_wallet(self) -> Optional['Abstract_Wallet']:
        if self.win is not None:
            if hasattr(self.win, 'wallet'):
                return self.win.wallet

    def get_gui_thread(self) -> Optional['threading.Thread']:
        if self.win is not None:
            if hasattr(self.win, 'gui_thread'):
                return self.win.gui_thread

    def update_status(self, paired: bool) -> None:
        pass

    def query_choice(self, msg: str, choices: Sequence[ChoiceItem]) -> Optional[Any]:
        """Returns ChoiceItem.key (for selected item), or None if the user cancels the dialog."""
        raise NotImplementedError()

    def yes_no_question(self, msg: str) -> bool:
        raise NotImplementedError()

    def show_message(self, msg: str, on_cancel=None) -> None:
        raise NotImplementedError()

    def show_error(self, msg: str, blocking: bool = False) -> None:
        raise NotImplementedError()

    def finished(self) -> None:
        pass

    def get_word(self, msg: str) -> str:
        raise NotImplementedError()

    def get_passphrase(self, msg: str, confirm: bool) -> Optional[str]:
        raise NotImplementedError()

    def get_pin(self, msg: str, *, show_strength: bool = True) -> str:
        raise NotImplementedError()


def is_any_tx_output_on_change_branch(tx: PartialTransaction) -> bool:
    return any([txout.is_change for txout in tx.outputs()])


def trezor_validate_op_return_output_and_get_data(output: TxOutput) -> bytes:
    validate_op_return_output(output)
    script = output.scriptpubkey
    if not (script[0] == opcodes.OP_RETURN and
            script[1] == len(script) - 2 and script[1] <= 75):
        raise UserFacingException(_("Only OP_RETURN scripts, with one constant push, are supported."))
    return script[2:]


def validate_op_return_output(output: TxOutput, *, max_size: int = None) -> None:
    script = output.scriptpubkey
    if script[0] != opcodes.OP_RETURN:
        raise UserFacingException(_("Only OP_RETURN scripts are supported."))
    if max_size is not None and len(script) > max_size:
        raise UserFacingException(_("OP_RETURN payload too large." + "\n"
                                  + f"(scriptpubkey size {len(script)} > {max_size})"))
    if output.value != 0:
        raise UserFacingException(_("Amount for OP_RETURN output must be zero."))


def only_hook_if_libraries_available(func):
    # note: this decorator must wrap @hook, not the other way around,
    # as 'hook' uses the name of the function it wraps
    def wrapper(self: 'HW_PluginBase', *args, **kwargs):
        if not self.libraries_available: return None
        return func(self, *args, **kwargs)
    return wrapper


class LibraryFoundButUnusable(Exception):
    def __init__(self, library_version='unknown'):
        self.library_version = library_version


class OutdatedHwFirmwareException(UserFacingException):

    def text_ignore_old_fw_and_continue(self) -> str:
        suffix = (_("The firmware of your hardware device is too old. "
                    "If possible, you should upgrade it. "
                    "You can ignore this error and try to continue, however things are likely to break.") + "\n\n" +
                  _("Ignore and continue?"))
        if str(self):
            return str(self) + "\n\n" + suffix
        else:
            return suffix


class OperationCancelled(UserFacingException):
    """Emitted when an operation is cancelled by user on a HW device
    """
    pass



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2015 Thomas Voegtlin
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# This module uses code from TLSLlite
# TLSLite Author: Trevor Perrin)


import binascii


def a2b_base64(s):
    try:
        b = bytearray(binascii.a2b_base64(s))
    except Exception as e:
        raise SyntaxError("base64 error: %s" % e)
    return b


def dePem(s, name):
    """Decode a PEM string into a bytearray of its payload.

    The input must contain an appropriate PEM prefix and postfix
    based on the input name string, e.g. for name="CERTIFICATE":

    -----BEGIN CERTIFICATE-----
    MIIBXDCCAUSgAwIBAgIBADANBgkqhkiG9w0BAQUFADAPMQ0wCwYDVQQDEwRUQUNL
    ...
    KoZIhvcNAQEFBQADAwA5kw==
    -----END CERTIFICATE-----

    The first such PEM block in the input will be found, and its
    payload will be base64 decoded and returned.
    """
    prefix  = "-----BEGIN %s-----" % name
    postfix = "-----END %s-----" % name
    start = s.find(prefix)
    if start == -1:
        raise SyntaxError("Missing PEM prefix")
    end = s.find(postfix, start+len(prefix))
    if end == -1:
        raise SyntaxError("Missing PEM postfix")
    s = s[start+len("-----BEGIN %s-----" % name) : end]
    retBytes = a2b_base64(s) # May raise SyntaxError
    return retBytes

def dePemList(s, name):
    """Decode a sequence of PEM blocks into a list of bytearrays.

    The input must contain any number of PEM blocks, each with the appropriate
    PEM prefix and postfix based on the input name string, e.g. for
    name="TACK BREAK SIG".  Arbitrary text can appear between and before and
    after the PEM blocks.  For example:

    " Created by TACK.py 0.9.3 Created at 2012-02-01T00:30:10Z -----BEGIN TACK
    BREAK SIG-----
    ATKhrz5C6JHJW8BF5fLVrnQss6JnWVyEaC0p89LNhKPswvcC9/s6+vWLd9snYTUv
    YMEBdw69PUP8JB4AdqA3K6Ap0Fgd9SSTOECeAKOUAym8zcYaXUwpk0+WuPYa7Zmm
    SkbOlK4ywqt+amhWbg9txSGUwFO5tWUHT3QrnRlE/e3PeNFXLx5Bckg= -----END TACK
    BREAK SIG----- Created by TACK.py 0.9.3 Created at 2012-02-01T00:30:11Z
    -----BEGIN TACK BREAK SIG-----
    ATKhrz5C6JHJW8BF5fLVrnQss6JnWVyEaC0p89LNhKPswvcC9/s6+vWLd9snYTUv
    YMEBdw69PUP8JB4AdqA3K6BVCWfcjN36lx6JwxmZQncS6sww7DecFO/qjSePCxwM
    +kdDqX/9/183nmjx6bf0ewhPXkA0nVXsDYZaydN8rJU1GaMlnjcIYxY= -----END TACK
    BREAK SIG----- "

    All such PEM blocks will be found, decoded, and return in an ordered list
    of bytearrays, which may have zero elements if not PEM blocks are found.
     """
    bList = []
    prefix  = "-----BEGIN %s-----" % name
    postfix = "-----END %s-----" % name
    while 1:
        start = s.find(prefix)
        if start == -1:
            return bList
        end = s.find(postfix, start+len(prefix))
        if end == -1:
            raise SyntaxError("Missing PEM postfix")
        s2 = s[start+len(prefix) : end]
        retBytes = a2b_base64(s2) # May raise SyntaxError
        bList.append(retBytes)
        s = s[end+len(postfix) : ]




import os
import threading
from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QAbstractListModel, QModelIndex
from PyQt6.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject

from electrum.i18n import _
from electrum.logging import get_logger
from electrum.util import (
    WalletFileException, standardize_path, InvalidPassword, send_exception_to_crash_reporter, UserFacingException,
)
from electrum.plugin import run_hook
from electrum.lnchannel import ChannelState
from electrum.storage import StorageReadWriteError, WalletStorage
from electrum.wallet import Abstract_Wallet

from .auth import AuthMixin, auth_protect
from .qefx import QEFX
from .qewallet import QEWallet
from .qewizard import QENewWalletWizard, QEServerConnectWizard, QETermsOfUseWizard

if TYPE_CHECKING:
    from electrum.daemon import Daemon
    from electrum.plugin import Plugins


# wallet list model. supports both wallet basenames (wallet file basenames)
# and whole Wallet instances (loaded wallets)
from .util import check_password_strength


class QEWalletListModel(QAbstractListModel):
    _logger = get_logger(__name__)

    # define listmodel rolemap
    _ROLE_NAMES= ('name', 'path', 'active')
    _ROLE_KEYS = range(Qt.ItemDataRole.UserRole, Qt.ItemDataRole.UserRole + len(_ROLE_NAMES))
    _ROLE_MAP  = dict(zip(_ROLE_KEYS, [bytearray(x.encode()) for x in _ROLE_NAMES]))

    def __init__(self, daemon: 'Daemon', parent=None):
        QAbstractListModel.__init__(self, parent)
        self.daemon = daemon
        self._wallets = []
        self.reload()

    def rowCount(self, index):
        return len(self._wallets)

    def roleNames(self):
        return self._ROLE_MAP

    def data(self, index, role):
        (wallet_name, wallet_path) = self._wallets[index.row()]
        role_index = role - Qt.ItemDataRole.UserRole
        role_name = self._ROLE_NAMES[role_index]
        if role_name == 'name':
            return wallet_name
        if role_name == 'path':
            return wallet_path
        if role_name == 'active':
            return self.daemon.get_wallet(wallet_path) is not None

    @pyqtSlot()
    def reload(self):
        self._logger.debug('enumerating available wallets')
        self.beginResetModel()
        self._wallets = []
        self.endResetModel()

        available = []
        wallet_folder = os.path.dirname(self.daemon.config.get_wallet_path())
        with os.scandir(wallet_folder) as it:
            for i in it:
                if i.is_file():
                    available.append(i.path)
        for path in sorted(available):
            wallet = self.daemon.get_wallet(path)
            self.add_wallet(wallet_path=path)

    def add_wallet(self, wallet_path):
        self.beginInsertRows(QModelIndex(), len(self._wallets), len(self._wallets))
        wallet_name = os.path.basename(wallet_path)
        wallet_path = standardize_path(wallet_path)
        item = (wallet_name, wallet_path)
        self._wallets.append(item)
        self.endInsertRows()

    def remove_wallet(self, path):
        i = 0
        wallets = []
        remove = -1
        for wallet_name, wallet_path in self._wallets:
            if wallet_path == path:
                remove = i
            else:
                wallets.append((wallet_name, wallet_path))
            i += 1

        if remove >= 0:
            self.beginRemoveRows(QModelIndex(), remove, remove)
            self._wallets = wallets
            self.endRemoveRows()

    @pyqtSlot(str, result=bool)
    def wallet_name_exists(self, name):
        for wallet_name, wallet_path in self._wallets:
            if name == wallet_name:
                return True
        return False

    @pyqtSlot(str, result=str)
    def pathForName(self, name):
        for wallet_name, wallet_path in self._wallets:
            if name == wallet_name:
                return wallet_path
        return ''

    @pyqtSlot(str)
    def updateWallet(self, path):
        i = 0
        for wallet_name, wallet_path in self._wallets:
            if wallet_path == path:
                mi = self.createIndex(i, i)
                self.dataChanged.emit(mi, mi, self._ROLE_KEYS)
                return
            i += 1


class QEDaemon(AuthMixin, QObject):
    instance = None  # type: Optional[QEDaemon]

    _logger = get_logger(__name__)

    _available_wallets = None
    _current_wallet = None
    _new_wallet_wizard = None
    _terms_of_use_wizard = None
    _server_connect_wizard = None
    _path = None
    _name = None
    _use_single_password = False
    _password = None
    _loading = False

    _backendWalletLoaded = pyqtSignal([str], arguments=['password'])

    availableWalletsChanged = pyqtSignal()
    fxChanged = pyqtSignal()
    newWalletWizardChanged = pyqtSignal()
    termsOfUseWizardChanged = pyqtSignal()
    serverConnectWizardChanged = pyqtSignal()
    loadingChanged = pyqtSignal()
    requestNewPassword = pyqtSignal()

    walletLoaded = pyqtSignal([str, str], arguments=['name', 'path'])
    walletRequiresPassword = pyqtSignal([str, str], arguments=['name', 'path'])
    walletOpenError = pyqtSignal([str], arguments=["error"])
    walletDeleteError = pyqtSignal([str, str], arguments=['code', 'message'])
    walletRenameError = pyqtSignal([str], arguments=['message'])
    verifyMessageError = pyqtSignal([str], arguments=['error'])

    def __init__(self, daemon: 'Daemon', plugins: 'Plugins', parent=None):
        super().__init__(parent)
        if QEDaemon.instance:
            raise RuntimeError('There should only be one QEDaemon instance')
        QEDaemon.instance = self
        self.daemon = daemon
        self.plugins = plugins
        self.qefx = QEFX(daemon.fx, daemon.config)

        self._backendWalletLoaded.connect(self._on_backend_wallet_loaded)

    @pyqtSlot()
    def passwordValidityCheck(self):
        if not self._walletdb._validPassword:
            self.walletRequiresPassword.emit(self._name, self._path)

    @pyqtSlot()
    @pyqtSlot(str)
    @pyqtSlot(str, str)
    def loadWallet(self, path=None, password=None):
        if self._loading:
            return
        self._loading = True

        if path is None:
            self._path = self.daemon.config.get('wallet_path')  # command line -w option
            if self._path is None:
                self._path = self.daemon.config.CURRENT_WALLET
        else:
            self._path = path
        if self._path is None:
            self._loading = False
            return

        self.loadingChanged.emit()

        self._path = standardize_path(self._path)
        self._name = os.path.basename(self._path)

        self._logger.debug('load wallet ' + str(self._path))

        # password unification helper:
        # - if pw not given (None), try pw of current wallet.
        # - but "" empty str passwords are kept as-is, to open passwordless wallets
        if password is None:
            password = self._password

        # map explicit empty str password to None. the backend disallows empty str passwords.
        if password == '':
            password = None

        wallet_already_open = self.daemon.get_wallet(self._path)
        if wallet_already_open is not None:
            password = QEWallet.getInstanceFor(wallet_already_open).password

        def load_wallet_task():
            success = False
            try:
                local_password = password  # need this in local scope
                wallet = None
                try:
                    wallet = self.daemon.load_wallet(
                        self._path,
                        password=local_password,
                        upgrade=True,
                        # might have a keystore password, but unencrypted storage. we want to prompt for pw even then:
                        force_check_password=True,
                    )
                except InvalidPassword:
                    self.walletRequiresPassword.emit(self._name, self._path)
                except FileNotFoundError:
                    self.walletOpenError.emit(_('File not found') + f":\n{self._path}")
                except StorageReadWriteError:
                    self.walletOpenError.emit(_('Could not read/write file'))
                except WalletFileException as e:
                    self.walletOpenError.emit(_('Could not open wallet: {}').format(str(e)))
                    if e.should_report_crash:
                        send_exception_to_crash_reporter(e)

                if wallet is None:
                    return

                if self.daemon.config.WALLET_SHOULD_USE_SINGLE_PASSWORD:
                    self._use_single_password = self._update_password_for_directory_and_unlock_wallets(old_password=local_password, new_password=local_password)
                    if not self._use_single_password and self.daemon.config.WALLET_ANDROID_USE_BIOMETRIC_AUTHENTICATION:
                        # we need to disable biometric auth if the user creates wallets with different passwords as
                        # we only store one encrypted password which is not associated to a specific wallet
                        self._logger.warning(f"disabling biometric authentication, not in single password mode")
                        self.daemon.config.WALLET_ANDROID_USE_BIOMETRIC_AUTHENTICATION = False
                        self.daemon.config.WALLET_ANDROID_BIOMETRIC_AUTH_WRAPPED_WALLET_PASSWORD = ''
                        self.daemon.config.WALLET_ANDROID_BIOMETRIC_AUTH_ENCRYPTED_WRAP_KEY = ''
                    self._password = local_password
                    self.singlePasswordChanged.emit()
                    self._logger.info(f'use single password: {self._use_single_password}')
                else:
                    self._logger.info('use single password disabled by config')
                self.daemon.config.WALLET_DID_USE_SINGLE_PASSWORD = self._use_single_password

                run_hook('load_wallet', wallet)

                success = True
                self._backendWalletLoaded.emit(local_password)
            finally:
                if not success:  # if successful, _loading guard will be reset by _on_backend_wallet_loaded
                    self._loading = False
                    self.loadingChanged.emit()

        threading.Thread(target=load_wallet_task, daemon=False).start()

    @pyqtSlot()
    @pyqtSlot(str)
    def _on_backend_wallet_loaded(self, password=None):
        self._logger.debug('_on_backend_wallet_loaded')
        wallet = self.daemon.get_wallet(self._path)
        assert wallet is not None
        self._current_wallet = QEWallet.getInstanceFor(wallet)
        self.availableWallets.updateWallet(self._path)
        wallet.unlock(password or None)  # not conditional on wallet.requires_unlock in qml, as
        # the auth wrapper doesn't pass the entered password, but instead we rely on the password in memory
        self._loading = False
        self.loadingChanged.emit()
        self.walletLoaded.emit(self._name, self._path)

    @pyqtSlot(QEWallet)
    @pyqtSlot(QEWallet, bool)
    @pyqtSlot(QEWallet, bool, bool)
    def checkThenDeleteWallet(self, wallet, confirm_requests=False, confirm_balance=False):
        if wallet.wallet.lnworker:
            lnchannels = wallet.wallet.lnworker.get_channel_objects()
            if any([channel.get_state() != ChannelState.REDEEMED and not channel.is_backup() for channel in lnchannels.values()]):
                self.walletDeleteError.emit('unclosed_channels', _('There are still channels that are not fully closed'))
                return

        num_requests = len(wallet.wallet.get_unpaid_requests())
        if num_requests > 0 and not confirm_requests:
            self.walletDeleteError.emit('unpaid_requests', _('There are still unpaid requests. Really delete?'))
            return

        c, u, x = wallet.wallet.get_balance()
        if c+u+x > 0 and not wallet.wallet.is_watching_only() and not confirm_balance:
            self.walletDeleteError.emit('balance', _('There are still coins present in this wallet. Really delete?'))
            return

        self.delete_wallet(wallet)

    @auth_protect(message=_('Really delete this wallet?'))
    def delete_wallet(self, wallet):
        path = standardize_path(wallet.wallet.storage.get_path())
        self._logger.debug('deleting wallet with path %s' % path)
        self._current_wallet = None
        # TODO walletLoaded signal is confusing
        self.walletLoaded.emit(None, None)

        if not self.daemon.delete_wallet(path):
            self.walletDeleteError.emit('error', _('Problem deleting wallet'))
            return

        self.availableWallets.remove_wallet(path)

    def wallet_path_from_wallet_name(self, wallet_name: str) -> str:
        return os.path.join(self.daemon.config.get_datadir_wallet_path(), wallet_name)

    @pyqtSlot(str, result=bool)
    def isValidWalletName(self, wallet_name: str) -> bool:
        if not wallet_name:
            return False
        if self.availableWallets.wallet_name_exists(wallet_name):
            return False
        # note: we should probably restrict wallet names to be alphanumeric (plus underscore, etc)...
        # try to prevent sketchy path traversals:
        for forbidden_char in ("/", "\\", ):
            if forbidden_char in wallet_name:
                return False
        # note: wallet names starting with '.' are allowed as hidden wallets; they are not shown
        # in the wallet list unless active (see Wallets.qml) but can be created via the wizard.
        # disallow double '..*' filenames though.
        if wallet_name.startswith('..'):
            return False
        if os.path.basename(wallet_name) != wallet_name:  # '/foo/bar/' returns 'bar'
            return False
        wallet_path = self.wallet_path_from_wallet_name(wallet_name)
        # validate that the path looks sane to the filesystem:
        try:
            temp_storage = WalletStorage(wallet_path)
        except (StorageReadWriteError, WalletFileException):
            return False
        except Exception:
            self._logger.exception("")
            return False
        if temp_storage.file_exists():
            return False
        return True

    @pyqtSlot(str)
    def renameWallet(self, new_name: str):
        wallet = self._current_wallet
        assert wallet, "name change without wallet?"
        old_path = standardize_path(wallet.wallet.storage.get_path())
        wallet_dir = os.path.dirname(old_path)
        new_path = standardize_path(os.path.join(wallet_dir, new_name))
        if old_path == new_path:
            return
        self._current_wallet = None
        self.daemon.stop_wallet(old_path)
        try:
            self.daemon.rename_wallet_file(old_path, new_path)
        except Exception as e:
            self.walletRenameError.emit(_('Error renaming wallet:\n') + str(e))
        self.walletLoaded.emit(None, None)

    @pyqtProperty(bool, notify=loadingChanged)
    def loading(self):
        return self._loading

    @pyqtProperty(QEWallet, notify=walletLoaded)
    def currentWallet(self):
        return self._current_wallet

    @pyqtProperty(QEWalletListModel, notify=availableWalletsChanged)
    def availableWallets(self):
        if not self._available_wallets:
            self._available_wallets = QEWalletListModel(self.daemon)

        return self._available_wallets

    @pyqtProperty(QEFX, notify=fxChanged)
    def fx(self):
        return self.qefx

    @pyqtSlot(str, result=list)
    def getWalletsUnlockableWithPassword(self, password: str) -> list[str]:
        """
        Returns any wallet that can be unlocked with the given password.
        Can be used as fallback to unlock another wallet the user entered a
        password that doesn't work for the current wallet but might work for another one.
        """
        wallet_dir = os.path.dirname(self.daemon.config.get_wallet_path())
        _, _, wallet_paths_can_unlock = self.daemon.check_password_for_directory(
            old_password=password,
            new_password=None,
            wallet_dir=wallet_dir,
        )
        if not wallet_paths_can_unlock:
            return []
        self._logger.debug(f"getWalletsUnlockableWithPassword: can unlock {len(wallet_paths_can_unlock)} wallets")
        return [str(path) for path in wallet_paths_can_unlock]

    @pyqtSlot(str, result=int)
    def numWalletsWithPassword(self, password: str) -> int:
        """Returns the number of wallets that can be unlocked with the given password"""
        wallet_paths_can_unlock = self.getWalletsUnlockableWithPassword(password)
        return len(wallet_paths_can_unlock)

    singlePasswordChanged = pyqtSignal()
    @pyqtProperty(bool, notify=singlePasswordChanged)
    def singlePasswordEnabled(self):
        """
        singlePasswordEnabled is False if:
            a.) the user has no wallet (and password) yet
            b.) the user has wallets with different passwords (legacy)
            c.) all wallets are locked, we couldn't check yet if they all use the same password
            d.) we are on desktop where different passwords are allowed
        """
        return self._use_single_password

    @pyqtProperty(str, notify=singlePasswordChanged)
    def singlePassword(self):
        """
        self._password is also set to the last loaded wallet password if we WANT a single password,
        but don't actually have a single password yet. So singlePassword being set doesn't strictly
        mean all wallets use the same password.
        """
        return self._password

    @singlePassword.setter
    def singlePassword(self, password: str):
        assert password
        assert self.daemon.config.WALLET_SHOULD_USE_SINGLE_PASSWORD
        if self._password != password:
            self._password = password
            self.singlePasswordChanged.emit()

    @pyqtSlot(result=str)
    def suggestWalletName(self):
        # FIXME why not use util.get_new_wallet_name ?
        i = 1
        while self.availableWallets.wallet_name_exists(f'wallet_{i}'):
            i = i + 1
        return f'wallet_{i}'

    @pyqtSlot()
    @auth_protect(method='wallet_password_only')
    def startChangePassword(self):
        if self._use_single_password:
            self.requestNewPassword.emit()
        else:
            self.currentWallet.requestNewPassword.emit()

    @pyqtSlot(str, result=bool)
    def setPassword(self, password):
        assert self._use_single_password
        assert password
        if not self._update_password_for_directory_and_unlock_wallets(old_password=self._password, new_password=password):
            return False
        self._password = password
        return True

    def _update_password_for_directory_and_unlock_wallets(self, *, old_password, new_password):
        # note: this assumes all wallet files are in a single directory.
        # change wallet passwords:
        ret = self.daemon.update_password_for_directory(old_password=old_password, new_password=new_password)
        # If some wallets just had their password changed, they got "locked" by wallet.update_password().
        # If the password is not unified yet, other loaded wallets might still be unlocked.
        # restore the invariant that all loaded wallets in qml must be unlocked:
        for w in self.daemon.get_wallets().values():
            if not w.is_unlocked():
                w.unlock(new_password)
            assert w.is_unlocked()
        return ret

    @pyqtProperty(QENewWalletWizard, notify=newWalletWizardChanged)
    def newWalletWizard(self):
        if not self._new_wallet_wizard:
            self._new_wallet_wizard = QENewWalletWizard(self, self.plugins)

        return self._new_wallet_wizard

    @pyqtProperty(QEServerConnectWizard, notify=serverConnectWizardChanged)
    def serverConnectWizard(self):
        if not self._server_connect_wizard:
            self._server_connect_wizard = QEServerConnectWizard(self)

        return self._server_connect_wizard

    @pyqtProperty(QETermsOfUseWizard, notify=termsOfUseWizardChanged)
    def termsOfUseWizard(self):
        if not self._terms_of_use_wizard:
            self._terms_of_use_wizard = QETermsOfUseWizard(self)
        return self._terms_of_use_wizard

    @pyqtSlot()
    def startNetwork(self):
        self.daemon.start_network()

    @pyqtSlot(str, str, str, result=bool)
    def verifyMessage(self, address, message, signature):
        try:
            return Abstract_Wallet.verify_message(address=address, signature=signature, message=message)
        except UserFacingException as e:
            self.verifyMessageError.emit(str(e))
            return False

    @pyqtSlot(str, result=int)
    def passwordStrength(self, password):
        if len(password) == 0:
            return 0
        return check_password_strength(password)[0]



from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QAbstractListModel, QModelIndex, pyqtProperty, pyqtSignal, pyqtSlot

from electrum.lnchannel import ChannelState
from electrum.lnutil import LOCAL, REMOTE
from electrum.logging import get_logger
from electrum.util import Satoshis
from electrum.gui import messages

from electrum.gui.common_qt.util import qt_event_listener, QtEventListener

from .qetypes import QEAmount
from .qemodelfilter import QEFilterProxyModel

if TYPE_CHECKING:
    from electrum.wallet import Abstract_Wallet


class QEChannelListModel(QAbstractListModel, QtEventListener):
    _logger = get_logger(__name__)

    # define listmodel rolemap
    _ROLE_NAMES=('cid', 'state', 'state_code', 'initiator', 'capacity', 'can_send',
                 'can_receive', 'l_csv_delay', 'r_csv_delay', 'send_frozen', 'receive_frozen',
                 'type', 'node_id', 'node_alias', 'short_cid', 'funding_tx', 'is_trampoline',
                 'is_backup', 'is_imported', 'local_capacity', 'remote_capacity')
    _ROLE_KEYS = range(Qt.ItemDataRole.UserRole, Qt.ItemDataRole.UserRole + len(_ROLE_NAMES))
    _ROLE_MAP  = dict(zip(_ROLE_KEYS, [bytearray(x.encode()) for x in _ROLE_NAMES]))
    _ROLE_RMAP = dict(zip(_ROLE_NAMES, _ROLE_KEYS))

    _network_signal = pyqtSignal(str, object)

    def __init__(self, wallet: 'Abstract_Wallet', parent=None):
        super().__init__(parent)
        self.wallet = wallet
        self._channels = []

        self._fm_backups = None
        self._fm_nobackups = None

        self.initModel()

        # To avoid leaking references to "self" that prevent the
        # window from being GC-ed when closed, callbacks should be
        # methods of this class only, and specifically not be
        # partials, lambdas or methods of subobjects.  Hence...
        self.register_callbacks()
        self.destroyed.connect(lambda: self.on_destroy())

    @qt_event_listener
    def on_event_channel(self, wallet, channel):
        if wallet == self.wallet:
            self.on_channel_updated(channel)

    @qt_event_listener
    def on_event_channels_updated(self, wallet):
        if wallet == self.wallet:
            self.initModel()

    def on_destroy(self):
        self.unregister_callbacks()

    def rowCount(self, index):
        return len(self._channels)

    # also expose rowCount as a property
    countChanged = pyqtSignal()
    @pyqtProperty(int, notify=countChanged)
    def count(self):
        return len(self._channels)

    def roleNames(self):
        return self._ROLE_MAP

    def data(self, index, role):
        tx = self._channels[index.row()]
        role_index = role - Qt.ItemDataRole.UserRole
        value = tx[self._ROLE_NAMES[role_index]]
        if isinstance(value, (bool, list, int, str, QEAmount)) or value is None:
            return value
        if isinstance(value, Satoshis):
            return value.value
        return str(value)

    def clear(self):
        self.beginResetModel()
        self._channels = []
        self.endResetModel()

    def channel_to_model(self, lnc):
        lnworker = self.wallet.lnworker
        item = {
            'cid': lnc.channel_id.hex(),
            'node_id': lnc.node_id.hex(),
            'node_alias': lnworker.lnpeermgr.get_node_alias(lnc.node_id) or '',
            'short_cid': lnc.short_id_for_GUI(),
            'state': lnc.get_state_for_GUI(),
            'state_code': int(lnc.get_state()),
            'is_backup': lnc.is_backup(),
            'is_trampoline': lnworker.is_trampoline_peer(lnc.node_id),
            'capacity': QEAmount(amount_sat=lnc.get_capacity())
        }
        if lnc.is_backup():
            item['can_send'] = QEAmount()
            item['can_receive'] = QEAmount()
            item['local_capacity'] = QEAmount()
            item['remote_capacity'] = QEAmount()
            item['send_frozen'] = True
            item['receive_frozen'] = True
            item['is_imported'] = lnc.is_imported
        else:
            item['can_send'] = QEAmount(amount_msat=lnc.available_to_spend(LOCAL))
            item['can_receive'] = QEAmount(amount_msat=lnc.available_to_spend(REMOTE))
            item['local_capacity'] = QEAmount(amount_msat=lnc.balance(LOCAL))
            item['remote_capacity'] = QEAmount(amount_msat=lnc.balance(REMOTE))
            item['send_frozen'] = lnc.is_frozen_for_sending()
            item['receive_frozen'] = lnc.is_frozen_for_receiving()
            item['is_imported'] = False
        return item

    numOpenChannelsChanged = pyqtSignal()
    @pyqtProperty(int, notify=numOpenChannelsChanged)
    def numOpenChannels(self):
        return sum([1 if x['state_code'] == ChannelState.OPEN else 0 for x in self._channels])

    @pyqtSlot()
    def initModel(self):
        self._logger.debug('init_model')
        if not self.wallet.lnworker:
            self._logger.warning('lnworker should be defined')
            return

        channels = []

        lnchannels = self.wallet.lnworker.get_channel_objects()
        for channel in lnchannels.values():
            item = self.channel_to_model(channel)
            channels.append(item)

        # sort, for now simply by state
        def chan_sort_score(c):
            return c['state_code'] + (10 if c['is_backup'] else 0)
        channels.sort(key=chan_sort_score)

        self.clear()
        self.beginInsertRows(QModelIndex(), 0, len(channels) - 1)
        self._channels = channels
        self.endInsertRows()

        self.countChanged.emit()

    def on_channel_updated(self, channel):
        for i, c in enumerate(self._channels):
            if c['cid'] == channel.channel_id.hex():
                self.do_update(i, channel)
                break

    def do_update(self, modelindex, channel):
        self._logger.debug(f'updating our channel {channel.short_id_for_GUI()}')
        modelitem = self._channels[modelindex]
        modelitem.update(self.channel_to_model(channel))

        mi = self.createIndex(modelindex, 0)
        self.dataChanged.emit(mi, mi, self._ROLE_KEYS)
        self.numOpenChannelsChanged.emit()

    @pyqtSlot(str)
    def newChannel(self, cid):
        self._logger.debug('new channel with cid %s' % cid)
        lnchannels = self.wallet.lnworker.channels
        for channel in lnchannels.values():
            if cid == channel.channel_id.hex():
                item = self.channel_to_model(channel)
                self._logger.debug(item)
                self.beginInsertRows(QModelIndex(), 0, 0)
                self._channels.insert(0, item)
                self.endInsertRows()
                self.countChanged.emit()
                return

    @pyqtSlot(str)
    def removeChannel(self, cid):
        self._logger.debug('remove channel with cid %s' % cid)
        for i, channel in enumerate(self._channels):
            if cid == channel['cid']:
                self._logger.debug(cid)
                self.beginRemoveRows(QModelIndex(), i, i)
                self._channels.remove(channel)
                self.endRemoveRows()
                self.countChanged.emit()
                return

    def filterModel(self, role, match):
        _filterModel = QEFilterProxyModel(self, self)
        assert role in self._ROLE_RMAP
        _filterModel.setFilterRole(self._ROLE_RMAP[role])
        _filterModel.setFilterValue(match)
        return _filterModel

    @pyqtSlot(result=QEFilterProxyModel)
    def filterModelBackups(self):
        self._fm_backups = self.filterModel('is_backup', True)
        return self._fm_backups

    @pyqtSlot(result=QEFilterProxyModel)
    def filterModelNoBackups(self):
        self._fm_nobackups = self.filterModel('is_backup', False)
        return self._fm_nobackups

    @pyqtSlot(result=str)
    def lightningWarningMessage(self):
        return messages.MSG_LIGHTNING_WARNING



#!/usr/bin/env python
#
# Copyright (C) 2010 Laszlo Nagy (nagylzs)
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Taken from: https://code.activestate.com/recipes/577334-how-to-debug-deadlocked-multi-threaded-programs/


"""Stack tracer for multi-threaded applications.
Useful for debugging deadlocks and hangs.

Usage:
    import stacktracer
    stacktracer.trace_start("trace.html", interval=5)
    ...
    stacktracer.trace_stop()

This will create a file named "trace.html" showing the stack traces of all threads,
updated every 5 seconds.
"""

import os
import sys
import threading
import time
import traceback
from typing import Optional

# 3rd-party dependency:
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter


def _thread_from_id(ident) -> Optional[threading.Thread]:
    return threading._active.get(ident)


def stacktraces():
    """Taken from http://bzimmer.ziclix.com/2008/12/17/python-thread-dumps/"""
    code = []
    for thread_id, stack in sys._current_frames().items():
        thread = _thread_from_id(thread_id)
        code.append(f"\n# thread_id={thread_id}. thread={thread}")
        for filename, lineno, name, line in traceback.extract_stack(stack):
            code.append(f'File: "{filename}", line {lineno}, in {name}')
            if line:
                code.append("  %s" % (line.strip()))

    return highlight("\n".join(code), PythonLexer(), HtmlFormatter(
        full=False,
        # style="native",
        noclasses=True,
    ))


class TraceDumper(threading.Thread):
    """Dump stack traces into a given file periodically.

    # written by nagylzs
    """

    def __init__(self, fpath, interval, auto):
        """
        @param fpath: File path to output HTML (stack trace file)
        @param auto: Set flag (True) to update trace continuously.
            Clear flag (False) to update only if file not exists.
            (Then delete the file to force update.)
        @param interval: In seconds: how often to update the trace file.
        """
        assert (interval > 0.1)
        self.auto = auto
        self.interval = interval
        self.fpath = os.path.abspath(fpath)
        self.stop_requested = threading.Event()
        threading.Thread.__init__(self)

    def run(self):
        while not self.stop_requested.is_set():
            time.sleep(self.interval)
            if self.auto or not os.path.isfile(self.fpath):
                self.dump_stacktraces()

    def stop(self):
        self.stop_requested.set()
        self.join()
        try:
            if os.path.isfile(self.fpath):
                os.unlink(self.fpath)
        except OSError:
            pass

    def dump_stacktraces(self):
        with open(self.fpath, "w+") as fout:
            fout.write(stacktraces())


_tracer = None  # type: Optional[TraceDumper]


def trace_start(fpath, interval=5, *, auto=True):
    """Start tracing into the given file."""
    global _tracer
    if _tracer is None:
        _tracer = TraceDumper(fpath, interval, auto)
        _tracer.daemon = True
        _tracer.start()
    else:
        raise Exception("Already tracing to %s" % _tracer.fpath)


def trace_stop():
    """Stop tracing."""
    global _tracer
    if _tracer is None:
        raise Exception("Not tracing, cannot stop.")
    else:
        _tracer.stop()
        _tracer = None



from electrum.plugin import hook
from electrum.i18n import _
from electrum.util import print_stderr
from electrum.hw_wallet import CmdLineHandler

from .trezor import TrezorPlugin, PASSPHRASE_ON_DEVICE

class TrezorCmdLineHandler(CmdLineHandler):
    def __init__(self):
        self.passphrase_on_device = False
        super().__init__()

    def get_passphrase(self, msg, confirm):
        import getpass
        print_stderr(msg)
        if self.passphrase_on_device and self.yes_no_question(_('Enter passphrase on device?')):
            return PASSPHRASE_ON_DEVICE
        else:
            return getpass.getpass('')

class Plugin(TrezorPlugin):
    handler = CmdLineHandler()
    @hook
    def init_keystore(self, keystore):
        if not isinstance(keystore, self.keystore_class):
            return
        keystore.handler = self.handler

    def create_handler(self, window):
        return self.handler



# loosely based on
# http://trevorius.com/scrapbook/uncategorized/pyqt-custom-abstractitemmodel/

from PyQt6 import QtCore


class CustomNode:

    def __init__(self, model: 'CustomModel', data):
        self.model = model
        self._data = data
        self._children = []
        self._parent = None
        self._row = 0

    def get_data(self):
        return self._data

    def get_data_for_role(self, index, role):
        # define in child class
        raise NotImplementedError()

    def childCount(self):
        return len(self._children)

    def child(self, row):
        if row >= 0 and row < self.childCount():
            return self._children[row]

    def parent(self):
        return self._parent

    def row(self):
        return self._row

    def addChild(self, child):
        child._parent = self
        child._row = len(self._children)
        self._children.append(child)


class CustomModel(QtCore.QAbstractItemModel):

    def __init__(self, parent, columncount):
        QtCore.QAbstractItemModel.__init__(self, parent)
        self._root = CustomNode(self, None)
        self._columncount = columncount

    def rowCount(self, index):
        if index.isValid():
            return index.internalPointer().childCount()
        return self._root.childCount()

    def columnCount(self, index):
        return self._columncount

    def addChild(self, node, _parent):
        if not _parent or not _parent.isValid():
            parent = self._root
        else:
            parent = _parent.internalPointer()
        parent.addChild(self, node)

    def index(self, row, column, _parent=None):
        # Performance-critical function

        if not _parent or not _parent.isValid():
            parent = self._root
        else:
            parent = _parent.internalPointer()

        # Open-coded
        #   if not QtCore.QAbstractItemModel.hasIndex(self, row, column, _parent):
        # the implementation is equivalent but it's in C++,
        # so VM entries take up inordinate amounts of time (up to 25% of refresh()):
        if row < 0 or column < 0 or row >= self.rowCount(_parent) or column >= self._columncount:
            return QtCore.QModelIndex()

        child = parent.child(row)
        if child:
            return QtCore.QAbstractItemModel.createIndex(self, row, column, child)
        else:
            return QtCore.QModelIndex()

    def parent(self, index):
        if index.isValid():
            node = index.internalPointer()
            if node:
                p = node.parent()
                if p:
                    return QtCore.QAbstractItemModel.createIndex(self, p.row(), 0, p)
            else:
                return QtCore.QModelIndex()
        return QtCore.QModelIndex()

    def data(self, index, role):
        if not index.isValid():
            return None
        node = index.internalPointer()
        return node.get_data_for_role(index, role)



"""Module for lnurl-related functionality."""
# https://github.com/sipa/bech32/tree/master/ref/python
# https://github.com/lnbits/lnurl

import asyncio
import json
from typing import Callable, Optional, NamedTuple, Any, TYPE_CHECKING
import re
import urllib.parse

import aiohttp.client_exceptions

from electrum import segwit_addr, util
from electrum.segwit_addr import bech32_decode, Encoding, convertbits, bech32_encode
from electrum.bolt11 import BOLT11DecodeException, BOLT11EncodeException
from electrum.network import Network
from electrum.logging import get_logger
from electrum.i18n import _


_logger = get_logger(__name__)


SUPPORTED_LNURL_SCHEMES = ('lnurlp', 'lnurlw')


class LNURLError(Exception): pass

class UntrustedLNURLError(LNURLError):
    def __init__(self, message=""):
        # use if error messages are returned by the LNURL server,
        # some services could try to trick users into doing something
        # by sending a malicious error message
        if message:
            message = (
                f"{_('[DO NOT TRUST THIS MESSAGE]:')}\n"
                f"{util.error_text_str_to_safe_str(message)}"
            )
        super().__init__(message)


def decode_lnurl(lnurl: str) -> str:
    """Converts bech32 encoded lnurl to url."""
    decoded_bech32 = bech32_decode(
        lnurl, ignore_long_length=True
    )
    hrp = decoded_bech32.hrp
    data = decoded_bech32.data
    if decoded_bech32.encoding is None:
        raise BOLT11DecodeException("Bad bech32 checksum")
    if decoded_bech32.encoding != Encoding.BECH32:
        raise BOLT11DecodeException("Bad bech32 encoding: must be using vanilla BECH32")
    if not hrp.startswith("lnurl"):
        raise BOLT11DecodeException("Does not start with lnurl")
    data = convertbits(data, 5, 8, False)
    url = bytes(data).decode("utf-8")
    return url


def encode_lnurl(url: str) -> str:
    """Encode url to bech32 lnurl string."""
    try:
        url = url.encode("utf-8")
    except UnicodeError as e:
        raise BOLT11EncodeException("invalid url") from e
    bech32_data = convertbits(url, 8, 5, True)
    assert bech32_data
    lnurl = bech32_encode(
        encoding=segwit_addr.Encoding.BECH32, hrp="lnurl", data=bech32_data)
    return lnurl.upper()


def _is_url_safe_enough_for_lnurl(url: str) -> bool:
    u = urllib.parse.urlparse(url)
    if u.scheme.lower() == "https":
        return True
    if u.netloc.endswith(".onion"):
        return True
    return False


def _parse_lnurl_response_callback_url(lnurl_response: dict) -> str:
    try:
        callback_url = lnurl_response['callback']
    except KeyError as e:
        raise LNURLError(f"Missing 'callback' field in lnurl response.") from e
    if not _is_url_safe_enough_for_lnurl(callback_url):
        raise LNURLError(
            f"This lnurl callback_url looks unsafe. It must use 'https://' or '.onion' (found: {callback_url[:10]}...)")
    return callback_url


# payRequest
# https://github.com/lnurl/luds/blob/227f850b701e9ba893c080103c683273e2feb521/06.md
class LNURL6Data(NamedTuple):
    callback_url: str
    max_sendable_sat: int
    min_sendable_sat: int
    metadata_plaintext: str
    comment_allowed: int
    #tag: str = "payRequest"

# withdrawRequest
# https://github.com/lnurl/luds/blob/227f850b701e9ba893c080103c683273e2feb521/03.md
class LNURL3Data(NamedTuple):
    # The URL which LN SERVICE would accept a withdrawal Lightning invoice as query parameter
    callback_url: str
    # Random or non-random string to identify the user's LN WALLET when using the callback URL
    k1: str
    # A default withdrawal invoice description
    default_description: str
    # Min amount the user can withdraw from LN SERVICE, or 0
    min_withdrawable_sat: int
    # Max amount the user can withdraw from LN SERVICE,
    # or equal to minWithdrawable if the user has no choice over the amounts
    max_withdrawable_sat: int

LNURLData = LNURL6Data | LNURL3Data


async def _request_lnurl(url: str) -> dict:
    """Requests payment data from a lnurl."""
    if not _is_url_safe_enough_for_lnurl(url):
        raise LNURLError(f"This lnurl looks unsafe. It must use 'https://' or '.onion' (found: {url[:10]}...)")
    try:
        response_raw = await Network.async_send_http_on_proxy("get", url, timeout=10)
    except asyncio.TimeoutError as e:
        raise LNURLError("LNURL server did not reply in time.") from e
    except aiohttp.client_exceptions.ClientError as e:
        raise LNURLError(f"Client error: {e}") from e
    try:
        response = json.loads(response_raw)
    except json.JSONDecodeError:
        raise LNURLError(f"Invalid response from LNURL server")

    status = response.get("status")
    if status and status == "ERROR":
        raise UntrustedLNURLError(f"LNURL request encountered an error: {response.get('reason', '<missing reason>')}")
    return response


def _parse_lnurl6_response(lnurl_response: dict) -> LNURL6Data:
    # parse lnurl6 "metadata"
    metadata_plaintext = ""
    try:
        metadata_raw = lnurl_response["metadata"]
        metadata = json.loads(metadata_raw)
        for m in metadata:
            if m[0] == 'text/plain':
                metadata_plaintext = str(m[1])
    except Exception as e:
        raise LNURLError(
            f"Missing or malformed 'metadata' field in lnurl6 response. exc: {e!r}") from e
    # parse lnurl6 "callback"
    callback_url = _parse_lnurl_response_callback_url(lnurl_response)
    # parse lnurl6 "minSendable"/"maxSendable"
    try:
        max_sendable_sat = int(lnurl_response['maxSendable']) // 1000
        min_sendable_sat = int(lnurl_response['minSendable']) // 1000
    except Exception as e:
        raise LNURLError(
            f"Missing or malformed 'minSendable'/'maxSendable' field in lnurl6 response. {e=!r}") from e
    # parse lnurl6 "commentAllowed" (optional, described in lnurl-12)
    try:
        comment_allowed = int(lnurl_response['commentAllowed']) if 'commentAllowed' in lnurl_response else 0
    except Exception as e:
        raise LNURLError(f"Malformed 'commentAllowed' field in lnurl6 response. {e=!r}") from e
    data = LNURL6Data(
        callback_url=callback_url,
        max_sendable_sat=max_sendable_sat,
        min_sendable_sat=min_sendable_sat,
        metadata_plaintext=metadata_plaintext,
        comment_allowed=comment_allowed,
    )
    return data


def _parse_lnurl3_response(lnurl_response: dict) -> LNURL3Data:
    """Parses the server response received when requesting a LNURL-withdraw (lud3) request"""
    callback_url = _parse_lnurl_response_callback_url(lnurl_response)
    if not (k1 := lnurl_response.get('k1')):
        raise UntrustedLNURLError(f"Missing k1 value in LNURL3 response: {lnurl_response=}")
    default_description = lnurl_response.get('defaultDescription', '')
    try:
        min_withdrawable_sat = int(lnurl_response['minWithdrawable'] or 0) // 1000
        max_withdrawable_sat = int(lnurl_response['maxWithdrawable']) // 1000
        assert max_withdrawable_sat >= min_withdrawable_sat, f"Invalid amounts: max < min amount"
        assert max_withdrawable_sat > 0, f"Invalid max amount: {max_withdrawable_sat} sat"
    except Exception as e:
        raise LNURLError(
            f"Missing or malformed 'minWithdrawable'/'minWithdrawable' field in lnurl3 response. {e=!r}") from e
    return LNURL3Data(
        callback_url=callback_url,
        k1=k1,
        default_description=default_description,
        min_withdrawable_sat=min_withdrawable_sat,
        max_withdrawable_sat=max_withdrawable_sat,
    )


async def request_lnurl(url: str) -> LNURLData:
    lnurl_dict = await _request_lnurl(url)
    tag = lnurl_dict.get('tag')
    if tag == 'payRequest':  # only LNURL6 is handled atm
        return _parse_lnurl6_response(lnurl_dict)
    elif tag == 'withdrawRequest':
        return _parse_lnurl3_response(lnurl_dict)
    raise UntrustedLNURLError(f"Unknown subtype of lnurl. tag={tag}")


async def try_resolve_lnurlpay(lnurl: Optional[str]) -> Optional[LNURL6Data]:
    if lnurl:
        try:
            result = await request_lnurl(lnurl)
            assert isinstance(result, LNURL6Data), f"lnurl result is not LNURL-pay response: {result=}"
            return result
        except Exception as request_error:
            _logger.debug(f"Error resolving lnurl: {request_error!r}")
    return None

async def request_lnurl_withdraw_callback(callback_url: str, k1: str, bolt_11: str) -> None:
    assert bolt_11
    params = {
        "k1": k1,
        "pr": bolt_11,
    }
    await callback_lnurl(
        url=callback_url,
        params=params
    )

async def callback_lnurl(url: str, params: dict) -> dict:
    """Requests an invoice from a lnurl supporting server."""
    if not _is_url_safe_enough_for_lnurl(url):
        raise LNURLError(f"This lnurl looks unsafe. It must use 'https://' or '.onion' (found: {url[:10]}...)")
    try:
        response_raw = await Network.async_send_http_on_proxy("get", url, params=params)
    except asyncio.TimeoutError as e:
        raise LNURLError("LNURL server did not reply in time.") from e
    except aiohttp.client_exceptions.ClientError as e:
        raise LNURLError(f"Client error: {e}") from e
    try:
        response = json.loads(response_raw)
        _logger.debug(f"lnurl response: {response}")
    except json.JSONDecodeError:
        raise LNURLError(f"Invalid response from LNURL server")

    status = response.get("status")
    if status and status == "ERROR":
        raise UntrustedLNURLError(f"LNURL request encountered an error: {response.get('reason', '<missing reason>')}")
    # TODO: implement LUD-09 (https://github.com/lnurl/luds/blob/luds/09.md), e.g. useful for paying offline devices
    if 'successAction' in response:
        raise LNURLError("successAction are not yet supported by Electrum.")
    # TODO: handling of specific errors (validate fields, e.g. for lnurl6)
    return response


LN_ADDRESS_RE = re.compile(
    r"[a-zA-Z0-9\-_][a-zA-Z0-9\-_.]*(\+[a-zA-Z0-9\-_.]+)?"          # username
    r"@"
    r"[a-zA-Z0-9\-]+(\.[a-zA-Z0-9\-]+)*"       # domain labels
    r"\.[a-zA-Z]{2,}"                          # tld (alphabetic, so no bare IPv4)
)


def lightning_address_to_url(address: str) -> Optional[str]:
    """Converts an email-type lightning address to a decoded lnurl.
    see https://github.com/fiatjaf/lnurl-rfc/blob/luds/16.md
    """
    if LN_ADDRESS_RE.fullmatch(address):
        username, domain = address.split("@")
        scheme = 'http' if domain.endswith('.onion') else 'https'
        return f"{scheme}://{domain}/.well-known/lnurlp/{username}"
